import numpy as np
import matplotlib.pyplot as plt

def getTimeAxis(precModel, timeUnits='s', bounds=None):
        '''
        Returns scaling factor, label and x-limits depending on units of time

        Parameters
        ----------
        timeUnits : str
            's' / 'sec' / 'seconds' - seconds
            'min' / 'minutes' - minutes
            'h' / 'hrs' / 'hours' - hours
        '''
        timeScale = 1
        timeLabel = 'Time (s)'
        if 'min' in timeUnits:
            timeScale = 1/60
            timeLabel = 'Time (min)'
        if 'h' in timeUnits:
            timeScale = 1/3600
            timeLabel = 'Time (hrs)'

        if bounds is None:
            bounds = [timeScale*1e-5*precModel.time[-1], timeScale * precModel.time[-1]]

        return timeScale, timeLabel, bounds

def plotBase(precModel, axes, variable, bounds = None, timeUnits = 's', radius='spherical', *args, **kwargs):
    '''
    Plots model outputs
    
    Parameters
    ----------
    axes : Axis
    variable : str
        Specified variable to plot
        Options are 'Volume Fraction', 'Total Volume Fraction', 'Critical Radius',
            'Average Radius', 'Volume Average Radius', 'Total Average Radius', 
            'Total Volume Average Radius', 'Aspect Ratio', 'Total Aspect Ratio'
            'Driving Force', 'Nucleation Rate', 'Total Nucleation Rate',
            'Precipitate Density', 'Total Precipitate Density', 
            'Temperature' and 'Composition'

            Note: for multi-phase simulations, adding the word 'Total' will
                sum the variable for all phases. Without the word 'Total', the variable
                for each phase will be plotted separately
                
    bounds : tuple (optional)
        Limits on the x-axis (float, float) or None (default, this will set bounds to (initial time, final time))
    timeUnits : str (optional)
        Plot time dependent variables per seconds ('s'), minutes ('min') or hours ('h')
    radius : str (optional)
        For non-spherical precipitates, plot the Average Radius by the -
            Equivalent spherical radius ('spherical')
            Short axis ('short')
            Long axis ('long')
        Note: Total Average Radius and Volume Average Radius will still use the equivalent spherical radius
    *args, **kwargs - extra arguments for plotting
    '''
    timeScale, timeLabel, bounds = getTimeAxis(precModel, timeUnits, bounds)

    axes.set_xlabel(timeLabel)
    axes.set_xlim(bounds)

    labels = {
        'Volume Fraction': 'Volume Fraction',
        'Total Volume Fraction': 'Volume Fraction',
        'Critical Radius': 'Critical Radius (m)',
        'Average Radius': 'Average Radius (m)',
        'Volume Average Radius': 'Volume Average Radius (m)',
        'Total Average Radius': 'Average Radius (m)',
        'Total Volume Average Radius': 'Volume Average Radius (m)',
        'Aspect Ratio': 'Mean Aspect Ratio',
        'Total Aspect Ratio': 'Mean Aspect Ratio',
        'Driving Force': 'Driving Force (J/m$^3$)',
        'Nucleation Rate': 'Nucleation Rate (#/m$^3$-s)',
        'Total Nucleation Rate': 'Nucleation Rate (#/m$^3$-s)',
        'Precipitate Density': 'Precipitate Density (#/m$^3$)',
        'Total Precipitate Density': 'Precipitate Density (#/m$^3$)',
        'Temperature': 'Temperature (K)',
        'Composition': 'Matrix Composition (at.%)',
        'Eq Composition Alpha': 'Matrix Composition (at.%)',
        'Eq Composition Beta': 'Matrix Composition (at.%)',
        'Supersaturation': 'Supersaturation',
        'Eq Volume Fraction': 'Volume Fraction'
    }

    totalVariables = ['Total Volume Fraction', 'Total Average Radius', 'Total Aspect Ratio', \
                        'Total Nucleation Rate', 'Total Precipitate Density', 'Total Volume Average Radius']
    singleVariables = ['Volume Fraction', 'Critical Radius', 'Average Radius', 'Aspect Ratio', \
                        'Driving Force', 'Nucleation Rate', 'Precipitate Density', 'Volume Average Radius']
    eqCompositions = ['Eq Composition Alpha', 'Eq Composition Beta']
    saturations = ['Supersaturation', 'Eq Volume Fraction']

    if variable == 'Temperature':
        plotTemperature(precModel, timeScale, labels, variable, axes, *args, **kwargs)
    elif variable == 'Composition':
        plotCompositions(precModel, timeScale, labels, variable, axes, *args, **kwargs)
    elif variable in eqCompositions:
        plotEqCompositions(precModel, timeScale, labels, variable, axes, *args, **kwargs)
    elif variable in saturations:
        plotSaurations(precModel, timeScale, labels, variable, axes, *args, **kwargs)
    elif variable in singleVariables:
        plotSingleVariables(precModel, timeScale, radius, labels, variable, axes, *args, **kwargs)
    elif variable in totalVariables:
        plotTotalVariables(precModel, timeScale, labels, variable, axes, *args, **kwargs)

def plotEuler(precModel, axes, variable, bounds = None, timeUnits = 's', radius='spherical', *args, **kwargs):
    '''
    Plots model outputs
    
    Parameters
    ----------
    axes : Axis
    variable : str
        Specified variable to plot
        Options are 'Volume Fraction', 'Total Volume Fraction', 'Critical Radius',
            'Average Radius', 'Volume Average Radius', 'Total Average Radius', 
            'Total Volume Average Radius', 'Aspect Ratio', 'Total Aspect Ratio'
            'Driving Force', 'Nucleation Rate', 'Total Nucleation Rate',
            'Precipitate Density', 'Total Precipitate Density', 
            'Temperature', 'Composition',
            'Size Distribution', 'Size Distribution Curve',
            'Size Distribution KDE', 'Size Distribution Density
            'Interfacial Composition Alpha', 'Interfacial Composition Beta'

            Note: for multi-phase simulations, adding the word 'Total' will
                sum the variable for all phases. Without the word 'Total', the variable
                for each phase will be plotted separately

                Interfacial composition terms are more relavent for binary systems than
                for multicomponent systems
                
    bounds : tuple (optional)
        Limits on the x-axis (float, float) or None (default, this will set bounds to (initial time, final time))
    radius : str (optional)
        For non-spherical precipitates, plot the Average Radius by the -
            Equivalent spherical radius ('spherical')
            Short axis ('short')
            Long axis ('long')
        Note: Total Average Radius and Volume Average Radius will still use the equivalent spherical radius
    *args, **kwargs - extra arguments for plotting
    '''
    sizeDistributionVariables = ['Size Distribution', 'Size Distribution Curve', 'Size Distribution KDE', 'Size Distribution Density']
    compositionVariables = ['Interfacial Composition Alpha', 'Interfacial Composition Beta']

    scale = []
    for p in range(len(precModel.phases)):
        if precModel.GB[p].nucleationSiteType == precModel.GB[p].BULK or precModel.GB[p].nucleationSiteType == precModel.GB[p].DISLOCATION:
            if radius == 'spherical':
                scale.append(precModel._GBareaRemoval(p) * np.ones(len(precModel.PBM[p].PSDbounds)))
            else:
                scale.append(1/precModel.shapeFactors[p].eqRadiusFactor(precModel.PBM[p].PSDbounds))
                if radius == 'long':
                    scale.append(precModel.shapeFactors[p].aspectRatio(precModel.PBM[p].PSDbounds) / precModel.shapeFactors[p].eqRadiusFactor(precModel.PBM[p].PSDbounds))
        else:
            scale.append(precModel._GBareaRemoval(p) * np.ones(len(precModel.PBM[p].PSDbounds)))

    if variable in compositionVariables:
        plotEulerComposition(precModel, variable, axes, *args, **kwargs)
    elif variable in sizeDistributionVariables:
        plotEulerSizeDistribution(precModel, scale, variable, axes, *args, **kwargs)
    elif variable == 'Cumulative Size Distribution':
        plotEulerCumulativeSizeDistribution(precModel, scale, variable, axes, *args, **kwargs)
    elif variable == 'Aspect Ratio Distribution':
        plotEulerAspectRatioDistribution(precModel, scale, variable, axes, *args, **kwargs) 
    else:
        plotBase(precModel, axes, variable, bounds, timeUnits, radius, *args, **kwargs)

def plotTemperature(precModel, timeScale, labels, variable, axes, *args, **kwargs):
    axes.semilogx(timeScale * precModel.time, precModel.temperature, *args, **kwargs)
    axes.set_ylabel(labels[variable])

def plotCompositions(precModel, timeScale, labels, variable, axes, *args, **kwargs):
    if precModel.numberOfElements == 1:
        axes.semilogx(timeScale * precModel.time, precModel.xComp[:,0], *args, **kwargs)
        axes.set_ylabel('Matrix Composition (at.% ' + precModel.elements[0] + ')')
    else:
        #If kwargs has label, add it as an extension to the label we add
        #And also pop label from kwargs so we don't have double arguments
        label_ext = ''
        if 'label' in kwargs:
            label_ext = '_' + kwargs['label']
            kwargs.pop('label')
        for i in range(precModel.numberOfElements):
            #Keep color consistent between Composition, Eq Composition Alpha and Eq Composition Beta if color isn't passed as an arguement
            if 'color' in kwargs:
                axes.semilogx(timeScale * precModel.time, precModel.xComp[:,i], label=precModel.elements[i] + label_ext, *args, **kwargs)
            else:
                axes.semilogx(timeScale * precModel.time, precModel.xComp[:,i], label=precModel.elements[i] + label_ext, color='C'+str(i), *args, **kwargs)
        axes.legend()
        axes.set_ylabel(labels[variable])
    yRange = [np.amin(precModel.xComp), np.amax(precModel.xComp)]
    axes.set_ylim([yRange[0] - 0.1 * (yRange[1] - yRange[0]), yRange[1] + 0.1 * (yRange[1] - yRange[0])])


def plotEqCompositions(precModel, timeScale, labels, variable, axes, *args, **kwargs):
    if variable == 'Eq Composition Alpha':
        plotVariable = precModel.xEqAlpha
    elif variable == 'Eq Composition Beta':
        plotVariable = precModel.xEqBeta

    if len(precModel.phases) == 1:
        if precModel.numberOfElements == 1:
            axes.semilogx(timeScale * precModel.time, plotVariable[:,0,0], *args, **kwargs)
            axes.set_ylabel('Matrix Composition (at.% ' + precModel.elements[0] + ')')
        else:
            for i in range(precModel.numberOfElements):
                #Keep color consistent between Composition, Eq Composition Alpha and Eq Composition Beta if color isn't passed as an arguement
                if 'color' in kwargs:
                    axes.semilogx(timeScale * precModel.time, plotVariable[:,0,i], label=precModel.elements[i]+'_Eq', *args, **kwargs)
                else:
                    axes.semilogx(timeScale * precModel.time, plotVariable[:,0,i], label=precModel.elements[i]+'_Eq', color='C'+str(i), *args, **kwargs)
            axes.legend()
            axes.set_ylabel(labels[variable])
    else:
        if precModel.numberOfElements == 1:
            for p in range(len(precModel.phases)):
                #Keep color somewhat consistent between Composition, Eq Composition Alpha and Eq Composition Beta if color isn't passed as an arguement
                if 'color' in kwargs:
                    axes.semilogx(timeScale * precModel.time, plotVariable[:,p,0], label=precModel.phases[p]+'_Eq', *args, **kwargs)
                else:
                    axes.semilogx(timeScale * precModel.time, plotVariable[:,p,0], label=precModel.phases[p]+'_Eq', color='C'+str(p), *args, **kwargs)
            axes.legend()
            axes.set_ylabel('Matrix Composition (at.% ' + precModel.elements[0] + ')')
        else:
            cIndex = 0
            for p in range(len(precModel.phases)):
                for i in range(precModel.numberOfElements):
                    #Keep color somewhat consistent between Composition, Eq Composition Alpha and Eq Composition Beta if color isn't passed as an arguement
                    if 'color' in kwargs:
                        axes.semilogx(timeScale * precModel.time, plotVariable[:,p,i], label=precModel.phases[p]+'_'+precModel.elements[i]+'_Eq', *args, **kwargs)
                    else:
                        axes.semilogx(timeScale * precModel.time, plotVariable[:,p,i], label=precModel.phases[p]+'_'+precModel.elements[i]+'_Eq', color='C'+str(cIndex), *args, **kwargs)
                    cIndex += 1
            axes.legend()
            axes.set_ylabel(labels[variable])

def plotSaurations(precModel, timeScale, labels, variable, axes, *args, **kwargs):
    #Since supersaturation is calculated in respect to the tie-line, it is the same for each element
    #Thus only a single element is needed
    plotVariable = np.zeros(precModel.betaFrac.shape)
    for p in range(len(precModel.phases)):
        if variable == 'Eq Volume Fraction':
            num = precModel.xComp[0,0] - precModel.xEqAlpha[:,p,0]
        else:
            num = precModel.xComp[:,0] - precModel.xEqAlpha[:,p,0]
        den = precModel.xEqBeta[:,p,0] - precModel.xEqAlpha[:,p,0]
        #If precipitate is unstable, both xEqAlpha and xEqBeta are set to 0
        #For these cases, change the values of numerator and denominator so that supersaturation is 0 instead of undefined
        num[den == 0] = 0
        den[den == 0] = 1
        plotVariable[:,p] = num / den
    
    if len(precModel.phases) == 1:
        axes.semilogx(timeScale * precModel.time, plotVariable[:,0], *args, **kwargs)
    else:
        for p in range(len(precModel.phases)):
            if 'color' in kwargs:
                axes.semilogx(timeScale * precModel.time, plotVariable[:,p], label=precModel.phases[p], *args, **kwargs)
            else:
                axes.semilogx(timeScale * precModel.time, plotVariable[:,p], label=precModel.phases[p], color='C'+str(p), *args, **kwargs)
        axes.legend()
    axes.set_ylabel(labels[variable])

def plotSingleVariables(precModel, timeScale, radius, labels, variable, axes, *args, **kwargs):
    if variable == 'Volume Fraction':
        plotVariable = precModel.betaFrac
    elif variable == 'Critical Radius':
        plotVariable = precModel.Rcrit
    elif variable == 'Average Radius':
        plotVariable = precModel.avgR
        for p in range(len(precModel.phases)):
            if precModel.GB[p].nucleationSiteType == precModel.GB[p].BULK or precModel.GB[p].nucleationSiteType == precModel.GB[p].DISLOCATION:
                if radius != 'spherical':
                    plotVariable[p] /= precModel.shapeFactors[p].eqRadiusFactor(precModel.avgR[p])
                if radius == 'long':
                    plotVariable[p] *= precModel.avgAR[p]
            else:
                plotVariable[p] *= precModel._GBareaRemoval(p)
    elif variable == 'Volume Average Radius':
        plotVariable = np.zeros(precModel.betaFrac.shape)
        indices = precModel.precipitateDensity > 0
        plotVariable[indices] = np.cbrt(precModel.betaFrac[indices] / precModel.precipitateDensity[indices] / (4/3*np.pi))
    elif variable == 'Aspect Ratio':
        plotVariable = precModel.avgAR
    elif variable == 'Driving Force':
        plotVariable = precModel.dGs
    elif variable == 'Nucleation Rate':
        plotVariable = precModel.nucRate
    elif variable == 'Precipitate Density':
        plotVariable = precModel.precipitateDensity

    if (len(precModel.phases)) == 1:
        axes.semilogx(timeScale * precModel.time, plotVariable[:,0], *args, **kwargs)
    else:
        for p in range(len(precModel.phases)):
            axes.semilogx(timeScale * precModel.time, plotVariable[:,p], label=precModel.phases[p], color='C'+str(p), *args, **kwargs)
        axes.legend()
    axes.set_ylabel(labels[variable])
    yb = 1 if variable == 'Aspect Ratio' else 0
    axes.set_ylim([yb, 1.1 * np.amax(plotVariable)])

def plotTotalVariables(precModel, timeScale, labels, variable, axes, *args, **kwargs):
    if variable == 'Total Volume Fraction':
        plotVariable = np.sum(precModel.betaFrac, axis=1)
    elif variable == 'Total Average Radius':
        totalN = np.sum(precModel.precipitateDensity, axis=1)
        totalN[totalN == 0] = 1
        totalR = np.sum(precModel.avgR * precModel.precipitateDensity, axis=1)
        plotVariable = totalR / totalN
    elif variable == 'Total Volume Average Radius':
        totalN = np.sum(precModel.precipitateDensity, axis=1)
        totalN[totalN == 0] = 1
        totalVol = np.sum(precModel.betaFrac, axis=1)
        plotVariable = np.cbrt(totalVol / totalN)
    elif variable == 'Total Aspect Ratio':
        totalN = np.sum(precModel.precipitateDensity, axis=1)
        totalN[totalN == 0] = 1
        totalAR = np.sum(precModel.avgAR * precModel.precipitateDensity, axis=1)
        plotVariable = totalAR / totalN
    elif variable == 'Total Nucleation Rate':
        plotVariable = np.sum(precModel.nucRate, axis=1)
    elif variable == 'Total Precipitate Density':
        plotVariable = np.sum(precModel.precipitateDensity, axis=1)

    axes.semilogx(timeScale * precModel.time, plotVariable, *args, **kwargs)
    axes.set_ylabel(labels[variable])
    yb = 1 if variable == 'Total Aspect Ratio' else 0
    axes.set_ylim(bottom=yb)

def plotEulerComposition(precModel, variable, axes, *args, **kwargs):
    if variable == 'Interfacial Composition Alpha':
        yVar = precModel.PSDXalpha
        ylabel = 'Composition in Alpha phase'
    else:
        yVar = precModel.PSDXbeta
        ylabel = 'Composition in Beta Phase'

    if (len(precModel.phases)) == 1:
        axes.semilogx(precModel.PBM[0].PSDbounds, yVar[0], *args, **kwargs)
    else:
        for p in range(len(precModel.phases)):
            axes.plot(precModel.PBM[p].PSDbounds, yVar[p], label=precModel.phases[p], *args, **kwargs)
        axes.legend()
    axes.set_xlim([precModel.PBM[0].PSDbounds[0], precModel.PBM[0].PSDbounds[-1]])
    axes.set_xlabel('Radius (m)')
    axes.set_ylabel(ylabel)

def plotEulerSizeDistribution(precModel, scale, variable, axes, *args, **kwargs):
    ylabel = 'Frequency (#/$m^3$)'
    if variable == 'Size Distribution':
        functionName = 'PlotHistogram'
    elif variable == 'Size Distribution KDE':
        functionName = 'PlotKDE'
    elif variable == 'Size Distribution Density':
        functionName = 'PlotDistributionDensity'
        ylabel = 'Distribution Density (#/$m^4$)'
    else:
        functionName = 'PlotCurve'

    if len(precModel.phases) == 1:
        getattr(precModel.PBM[0], functionName)(axes, scale=scale[0], *args, **kwargs)
    else:
        for p in range(len(precModel.phases)):
            getattr(precModel.PBM[p], functionName)(axes, label=precModel.phases[p], scale=scale[p], *args, **kwargs)
        axes.legend()
    axes.set_xlabel('Radius (m)')
    axes.set_ylabel(ylabel)
    axes.set_xlim([0, np.amax([pb.max for pb in precModel.PBM])])
    if variable == 'Size Distribution Density':
        axes.set_ylim([0, 1.1*np.amax(np.concatenate(([np.amax(pb.PSD/(pb.PSDbounds[1:] - pb.PSDbounds[:-1])) for pb in precModel.PBM], [1])))])
    else:
        axes.set_ylim([0, 1.1*np.amax(np.concatenate(([np.amax(pb.PSD) for pb in precModel.PBM], [1])))])

def plotEulerCumulativeSizeDistribution(precModel, scale, variable, axes, *args, **kwargs):
    ylabel = 'CDF'
    if len(precModel.phases) == 1:
        precModel.PBM[0].PlotCDF(axes, scale=scale[0], *args, **kwargs)
    else:
        for p in range(len(precModel.phases)):
            precModel.PBM[p].PlotCDF(axes, label=precModel.phases[p], scale=scale[p], *args, **kwargs)
        axes.legend()
    axes.set_xlabel('Radius (m)')
    axes.set_ylabel(ylabel)
    axes.set_xlim([0, np.amax([pb.max for pb in precModel.PBM])])

def plotEulerAspectRatioDistribution(precModel, scale, variable, axes, *args, **kwargs):
    if len(precModel.phases) == 1:
        axes.plot(precModel.PBM[0].PSDbounds * np.interp(precModel.PBM[0].PSDbounds, precModel.PBM[0].PSDbounds, scale[0]), precModel.eqAspectRatio[0], *args, **kwargs)
    else:
        for p in range(len(precModel.phases)):
            axes.plot(precModel.PBM[p].PSDbounds * np.interp(precModel.PBM[p].PSDbounds, precModel.PBM[p].PSDbounds, scale[p]), precModel.eqAspectRatio[p], label=precModel.phases[p], *args, **kwargs)
        axes.legend()
    axes.set_xlim([0, np.amax([precModel.PBM[p].PSDbounds * np.interp(precModel.PBM[p].PSDbounds, precModel.PBM[p].PSDbounds, scale[p]) for p in range(len(precModel.phases))])])
    axes.set_ylim(bottom=1)
    axes.set_xlabel('Radius (m)')
    axes.set_ylabel('Aspect ratio distribution')




from .Thermodynamics import GeneralThermodynamics
from .BinTherm import BinaryThermodynamics
from .MultiTherm import MulticomponentThermodynamics
from .Surrogate import BinarySurrogate, MulticomponentSurrogate, generateTrainingPoints


'''
Built-in iterators

Currently, this is explicit-euler and 4th order runga kutta
''' 
def ExplicitEulerIterator(f, t, X_old, updateX):
    '''
    Explicit euler iteration scheme

    Defined by:
        dXdt = f(t, X_n)
        X_n+1 = X_n + f(t, X_n) * dt

    Parameters
    ----------
    f : function
        dX/dt - function taking in time and X and returning dX/dt
    t : float
        Current time
    X_old : list of arrays
        X at time t
    updateX : function
        Helper function to handle any correction to dxdt
        Takes in X_old, dxdt, dt and returns X_new

    Returns
    -------
    X_new : unformatted list of floats
        New values of X in format of X_old
    dt : float
        Time step
    '''
    dxdt, dt = f(t, X_old, True)
    return updateX(X_old, dxdt, dt), dt

def RK4Iterator(f, t, X_old, updateX):
    '''
    4th order Runga Kutta iteration scheme

    Defined by:
        k1 = f(t, X_n)
        k2 = f(t + dt/2, X_n + k1 * dt/2)
        k3 = f(t + dt/2, X_n + k2 * dt/2)
        k4 = f(t + dt, X_n, k3 * dt)
        X_n+1 = X_n + 1/6 * (k1 + 2*k2 + 2*k3 + k4) * dt

    Parameters
    ----------
    f : function
        dX/dt - function taking in time and X and returning dX/dt
    t : float
        Current time
    X_old : list of arrays
        X at time t
    updateX : function
        Helper function to handle any correction to dxdt
        Takes in X_old, dxdt, dt and returns X_new

    Returns
    -------
    X_new : unformatted list of floats
        New values of X in format of X_old
    dt : float
        Time step, important if modified from dtfunc
    '''
    dxdt, dt = f(t, X_old, True)

    k1 = dxdt
    dxdtsum = k1
    X_k1 = updateX(X_old, k1, dt/2)

    k2 = f(t, X_k1)
    dxdtsum += 2*k2
    X_k2 = updateX(X_old, k2, dt/2)

    k3 = f(t, X_k2)
    dxdtsum += 2*k3
    X_k3 = updateX(X_old, k3, dt)

    k4 = f(t, X_k3)
    dxdtsum += k4

    return updateX(X_old, dxdtsum/6, dt), dt


import numpy as np
from kawin.diffusion.Diffusion import DiffusionModel

class SinglePhaseModel(DiffusionModel):
    def _getFluxes(self, t, x_curr):
        '''
        Private function that gets fluxes at the boundary of each nodes given an array of compositions and current time

        Steps:
            1. Get diffusivity from cell centers using cell compositions
            2. Interpolate diffusivity to get diffusivity (D) at cell boundaries
            3. Calculate fluxes from concentration gradient (dx/dz) and interpolated diffusivity = -D * dx/dz
            4. Apply boundary conditions for fluxes at ends of mesh
                If fixed flux condition (Neumann) - then use the flux defined in the condition
                If fixed composition condition (Dirichlet) - then use nearby flux (this will keep the composition fixed after apply the fluxes)
            5. Store dt (from von Neumann analysis) for later

        Returns
        -------
        fluxes : (e-1, n+1) array of floats
            e - number of elements including reference element
            n - number of nodes
        dt : float
            Maximum calculated time interval for numerical stability
        '''
        #Calculate diffusivity at cell centers
        x = x_curr[0]
        T = self.Tfunc(self.z, t)
        if len(self.elements) == 1:
            d = np.zeros(self.N)
        else:
            d = np.zeros((self.N, len(self.elements), len(self.elements)))
        if self.cache:
            for i in range(self.N):
                hashValue = self._getHash(x[:,i], T[i])
                if hashValue not in self.hashTable:
                    self.hashTable[hashValue] = self.therm.getInterdiffusivity(x[:,i], T[i], phase=self.phases[0])
                d[i] = self.hashTable[hashValue]
        else:
            d = self.therm.getInterdiffusivity(x.T, T, phase=self.phases[0])
        
        #Get diffusivity and composition gradient at cell boundaries
        dmid = (d[1:] + d[:-1]) / 2
        dxdz = (x[:,1:] - x[:,:-1]) / self.dz

        #Fluxes = -D * dx/dz
        fluxes = np.zeros((len(self.elements), self.N+1))
        if len(self.elements) == 1:
            fluxes[0,1:-1] = -dmid * dxdz
        else:
            dxdz = np.expand_dims(dxdz, axis=0)
            fluxes[:,1:-1] = -np.matmul(dmid, np.transpose(dxdz, (2,1,0)))[:,:,0].T

        #Boundary condition
        for e in range(len(self.elements)):
            fluxes[e,0] = self.LBCvalue[e] if self.LBC[e] == self.FLUX else fluxes[e,1]
            fluxes[e,-1] = self.RBCvalue[e] if self.RBC[e] == self.FLUX else fluxes[e,-2]

        #Time step from von Neumann analysis (using 0.4 instead of 0.5 to be safe)
        self._currdt = 0.4 * self.dz**2 / np.amax(np.abs(dmid))

        return fluxes

    def getFluxes(self):
        '''
        Gets fluxes at the boundary of each nodes

        This calls the private _getFluxes method with the internal current x and t

        Returns
        -------
        fluxes : (e-1, n+1) array of floats
            e - number of elements including reference element
            n - number of nodes
        dt : float
            Maximum calculated time interval for numerical stability
        '''
        fluxes = self._getFluxes(self.t, [self.x])
        dt = self._currdt
        return fluxes, dt
    
    def getDt(self, dXdt):
        '''
        Returns dt that was calculated from _getFluxes
        This prevents double calculation of the diffusivity just to get a time step
        '''
        return self._currdt


import numpy as np
import scipy.stats as sts
import matplotlib.pyplot as plt
import copy

class PopulationBalanceModel:
    '''
    Class for handling particle size distributions (PSD)
        This include time evolution, moments and graphing

    NOTE: more formally, the PSD is defined such that the number of particles of size R in a volume
        is the integral of the PSD from R-1/2 to R+1/2
        For the discrete PSD, n_i = PSD_i * (R+1/2 - R-1/2)
        Here, we just store the PSD as the number of particles (n_i), but the formulation for 
        growth rates and moments are the same since the factor (R+1/2 - R-1/2) cancels out

    Parameters
    ----------
    cMin : float (optional)
        Lower bound of PSD (defaults at 1e-10)
    cMax : float (optional)
        Upper bound of PSD (defaults at 1e-9)
    bins : int (optional)
        Initial number of bins (defaults at 150)
    minBins : int (optional)
        Minimum number of bins over an order of magnitude (defaults at 100)
    maxBins : int (optional)
        Maximum number of bins over an order of magnitude (defaults at 200)

    Attributes
    ----------
    originalMin : float
        Minimum bin size initially set
    min : float
        Current minimum bin size (this will almost always be equal to originalMin)
    originalMax : float
        Maximum bin size initially set
    max : float
        Current maximum bin size
    bins : int
        Default number of bins
    minBins : int
        Minimum number of allowed bins
    maxBins : int
        Maximum number of allowed bins
    PSD : array
        Particle size distribution
    PSDsize : array
        Average radius of each PSD size class
    PSDbounds : array
        Radius at the bounds of each size class - length of array is len(PSD)+1
    '''
    def __init__(self, cMin = 1e-10, cMax = 1e-9, bins = 150, minBins = 100, maxBins = 200):
        self.originalMin = cMin
        self.originalMax = np.amax([10*self.originalMin, cMax])
        self.min = self.originalMin
        self.max = self.originalMax

        self.originalBins = bins
        self.setBinConstraints(bins, minBins, maxBins)
        
        self.reset()

        self._adaptiveBinSize = True
        
        self._record = False
        self._recordedBins = None
        self._recordedPSD = None
        self._recordedTime = None

    def reset(self, resetBounds = True):
        '''
        Resets the PSD to 0 and resets bin size and number of bins to original values
        This will remove any size classes that were added since initialization
        '''
        if resetBounds:
            self.min = self.originalMin
            self.max = self.originalMax
            self.bins = self.originalBins
        self.PSDbounds = np.linspace(self.min, self.max, self.bins+1)
        self.PSDsize = 0.5 * (self.PSDbounds[:-1] + self.PSDbounds[1:])
            
        self.PSD = np.zeros(self.bins)

        #Hidden variable for use in KWNEuler when adaptive time stepping is enabled
        #This allows for PSD to revert to its previous value if a time constraint is not met
        self._prevPSD = np.zeros(self.bins)
        self._prevPSDbounds = np.zeros(self.bins+1)

        #Temporary storage for net flux
        #This is used to correct the fluxes once the time step is known
        self._netFlux = None

    def enableRecording(self):
        '''
        Enables recording of particle size distribution per iteration

        The initial data in the recorded bin is t = 0, N_i = 0

        The size of the recorded particle size distribution will be (n x max bins)
            Where n in the number of iterations
            max bins is the maximum number of bins, if the current number is smaller, the rest of the array will be 0
        '''
        self._record = True
        self._recordedBins = np.zeros((1, self.maxBins + 1))
        self._recordedPSD = np.zeros((1, self.maxBins))
        self._recordedTime = np.zeros(1)

    def resetRecordedData(self):
        '''
        If recording, then reset the recorded bins to the original size (starting with t = 0, N_i = 0)
        If not recording, then clear the recorded data
        '''
        if self._record:
            self._recordedBins = np.zeros((1, self.maxBins + 1))
            self._recordedPSD = np.zeros((1, self.maxBins))
            self._recordedTime = np.zeros(1)
        else:
            self._recordedBins = None
            self._recordedPSD = None
            self._recordedTime = None

    def disableRecording(self):
        '''
        Disables recording

        We won't clear the recorded bins here in case the user still wants to grab recorded data
        '''
        self._record = False

    def setRecording(self, record = True):
        '''
        Wrapper around enable and disable recording
        '''
        if record:
            self.enableRecording()
        else:
            self.disableRecording()

    def removeRecordedData(self):
        '''
        Removes recorded data
        '''
        self._recordedBins = None
        self._recordedPSD = None
        self._recordedTime = None

    def record(self, time):
        '''
        Adds current PSD data to recorded arrays

        TODO: Make sure this works when adaptive bins is False
        '''
        if self._record:
            maxBins = self.maxBins if self._adaptiveBinSize else self.bins
            self._recordedBins = np.pad(self._recordedBins, ((0, 1), (0, maxBins+1 - self._recordedBins.shape[1])))
            self._recordedPSD = np.pad(self._recordedPSD, ((0, 1), (0, maxBins - self._recordedPSD.shape[1])))
            self._recordedTime = np.pad(self._recordedTime, (0,1))
            self._recordedBins[-1][:self.PSDbounds.shape[0]] = self.PSDbounds
            self._recordedPSD[-1][:self.PSD.shape[0]] = self.PSD
            self._recordedTime[-1] = time

    def saveRecordedPSD(self, filename, compressed = True):
        '''
        Saves recorded data into npz format

        Note: If recording is disabled, then this function will do nothing since 
              there is nothing to save anyways

        Parameters
        ----------
        filename : str
            File name to save to
        compressed : bool (optional)
            Whether to save as in compressed format (defaults to True)
        '''
        if self._record:
            if compressed:
                np.savez_compressed(filename, time = self._recordedTime, bins = self._recordedBins, PSD = self._recordedPSD)
            else:
                np.savez(filename, time = self._recordedTime, bins = self._recordedBins, PSD = self._recordedPSD)

    def loadRecordedPSD(self, filename):
        '''
        Loads recorded PSD
        '''
        data = np.load(filename)
        self._record = True
        self._recordedTime = data['time']
        self._recordedBins = data['bins']
        self._recordedPSD = data['PSD']

    def _grabPSDfromIndex(self, index):
        '''
        Returns PSD bounds, PSD bins and PSD from recorded data based off index

        Since the number of bins is likely less than the max, we want to grab only the non-zero indices
        TODO: two concerns
            1) this may remove the last 1 bins (this may be okay since we add new bins once the
                list bins has at least 1 particle), so the last bin would be 0 anyways
        '''
        nonzero = len(np.nonzero(self._recordedBins[index])[0])
        if nonzero == 0:
            PSDbounds = np.linspace(self.originalMin, self.originalMax, self.originalBins+1)
            PSDsize = 0.5 * (PSDbounds[1:] + PSDbounds[:-1])
            PSD = np.zeros(self.originalBins)
        else:
            PSDbounds = self._recordedBins[index,:nonzero]
            PSD = self._recordedPSD[index,:nonzero-1]
            PSDsize = 0.5 * (PSDbounds[1:] + PSDbounds[:-1])
        bins = len(PSD)
        minBound, maxBound = np.amin(PSDbounds), np.amax(PSDbounds)
        return PSDbounds, PSD, PSDsize, bins, minBound, maxBound

    def setPSDtoRecordedTime(self, time):
        '''
        Sets particle size distribution to specific time if recorded

        Parameter
        ---------
        time : float
            Time to load PSD from, will load to nearest time available
        '''
        if self._record:
            if time <= self._recordedTime[0]:
                print('Input time is lower than smallest recorded time, setting PSD to t = {:.3e}'.format(self._recordedTime[0]))
                self.PSDbounds, self.PSD, self.PSDsize, self.bins, self.min, self.max = self._grabPSDfromIndex(0)
            elif time >= self._recordedTime[-1]:
                print('Input time is larger than longest recorded time, setting PSD to t = {:.3e}'.format(self._recordedTime[-1]))
                self.PSDbounds, self.PSD, self.PSDsize, self.bins, self.min, self.max = self._grabPSDfromIndex(-1)
            else:
                #Upper and lower PSD
                #Note: horrible naming convention here
                #    Upper PSD refers to the PSD just after time
                #    Lower PSD refers to the PSD just before time
                #This does NOT refer to the PSD with the larger or smaller number of bins
                uind = np.argmax(self._recordedTime > time)
                lind = uind - 1

                utime, ltime = self._recordedTime[uind], self._recordedTime[lind]
                uPSDbounds, uPSD, uPSDsize, ubins, umin, umax = self._grabPSDfromIndex(uind)
                lPSDbounds, lPSD, lPSDsize, lbins, lmin, lmax = self._grabPSDfromIndex(lind)

                #Interpolate from lower PSD to upper PSD using bounds of larger PSD
                #This will account for all possible cases if the PSD size classes change
                #This is done by pretending we're calling changeSizeClasses
                #    Where we resize the PSD with the smaller number of bins to have the same bins as the larger PSD
                #    And correct for the possible change in number density
                if ubins >= lbins:
                    #Resize lower PSD to upper PSD
                    oldV = np.sum(lPSD * lPSDsize**3)
                    distDen = lPSD / (lPSDbounds[1:] - lPSDbounds[:-1])
                    rOld = 0.5 * (lPSDbounds[1:] + lPSDbounds[:-1])
                    lPSD = np.interp(uPSDsize, rOld, distDen, left=0, right=0) * (uPSDbounds[1:] - uPSDbounds[:-1])
                    newV = np.sum(lPSD * uPSDsize**3)
                    if newV != 0:
                        lPSD *= oldV / newV
                    else:
                        lPSD = np.zeros(ubins)
                    
                else:
                    #Resize upper PSD to lower PSD
                    oldV = np.sum(uPSD * uPSDsize**3)
                    distDen = uPSD / (uPSDbounds[1:] - uPSDbounds[:-1])
                    rOld = 0.5 * (uPSDbounds[1:] + uPSDbounds[:-1])
                    uPSD = np.interp(lPSDsize, rOld, distDen, left=0, right=0) * (lPSDbounds[1:] - lPSDbounds[:-1])
                    uPSDbounds = lPSDbounds
                    newV = np.sum(uPSD * lPSDsize**3)
                    if newV != 0:
                        uPSD *= oldV / newV
                    else:
                        uPSD = np.zeros(lbins)

                #Now that the bin sizes are the same, we can just interpolate the PSD
                self.PSDbounds = uPSDbounds
                self.PSDsize = 0.5 * (self.PSDbounds[1:] + self.PSDbounds[:-1])
                self.PSD = (uPSD - lPSD) * (time - ltime) / (utime - ltime) + lPSD
                self.bins = len(self.PSDsize)
                self.min, self.max = np.amin(self.PSDbounds), np.amax(self.PSDbounds)

    def setAdaptiveBinSize(self, adaptive):
        '''
        For Euler implementation, sets whether to change the bin size when 
        the number of filled bins > maxBins or < minBins

        If False, the bins will still be if nucleated particles are greater than the max bin size
        and bins will still be added when the last bins starts to fill (but this will not change the bin size)
        '''
        self._adaptiveBinSize = adaptive

    def setBinConstraints(self, bins = 150, minBins = 100, maxBins = 200):
        '''
        Sets constraints for minimum and maxinum number of bins over an order of magnitude

        All bins will be overridden to an even number
        Minimum number of bins will be overridden to be at most half of maximum bins
        Default bins will be overridden to be halfway between min and max bins if out of range
        '''
        self.minBins = minBins
        self.maxBins = maxBins
        self.bins = bins

    def LoadDistribution(self, data):
        '''
        Creates a particle size distribution from a set of data
        
        Parameters
        ----------
        data : array of floats
            Array of data to be inserted into PSD
        '''
        self.PSD, self.PSDbounds = np.histogram(data, self.PSDbounds)
        self.PSD = self.PSD.astype('float')

    def LoadDistributionFunction(self, function):
        '''
        Creates a particle size distribution from a function

        Parameters
        ----------
        function : function
            Takes in R and returns density
        '''
        self.PSD = function(self.PSDsize)

    def createBackup(self):
        '''
        Stores current PSD and PSDbounds
        '''
        self._prevPSD = copy.copy(self.PSD)
        self._prevPSDbounds = copy.copy(self.PSDbounds)

    def revert(self):
        '''
        Reverts to previous PSD and PSDbounds

        NOTE: this appears to be unused
            (this was used in the previous KWNEuler implementation when the PSD could change within an iteration)
            (now it changes between iterations, so we don't need to revert back if something goes wrong)
        '''
        self.PSD = copy.copy(self._prevPSD)
        self.PSDbounds = copy.copy(self._prevPSDbounds)
        self.PSDsize = 0.5 * (self.PSDbounds[1:] + self.PSDbounds[:-1])
        self.bins = len(self.PSD)
        self.min, self.max = self.PSDbounds[0], self.PSDbounds[-1]

    def changeSizeClasses(self, cMin, cMax, bins = None, resetPSD = False):
        '''
        Changes the size classes and resets the PSD

        This is done by linear interpolation of the previous bins and PSD
        And interpolating to the new bins and PSD
        Due to differences in bin size (thus resolution of the PSD), the number density
            could be a little different. To correct for this, we get the 3rd moment of the
            previous PSD and the new PSD, and correct the new PSD to have the same 3rd moment
        
        Parameters
        ----------
        cMin : float
            Lower bound of PSD
        cMax : float
            Upper bound of PSD
        bins : int
            Number of bins
        resetPSD : bool (optional)
            Whether to reset the PSD (defaults to False)
        '''
        self.bins = self.bins if bins is None else bins
        self.min = cMin
        self.max = np.amax([10*self.min, cMax])

        if resetPSD:
            self.reset()
        else:
            oldV = self.ThirdMoment()
            distDen = self.PSD / (self.PSDbounds[1:] - self.PSDbounds[:-1])
            rOld = 0.5 * (self.PSDbounds[1:] + self.PSDbounds[:-1])
            self.reset(False)
            self.PSD = np.interp(self.PSDsize, rOld, distDen) * (self.PSDbounds[1:] - self.PSDbounds[:-1])
            newV = self.ThirdMoment()
            if newV != 0:
                self.PSD *= oldV / newV
            else:
                self.PSD = np.zeros(self.bins)

    def addSizeClasses(self, bins = 1):
        '''
        Adds an additional number of size classes to end of distribution

        Parameters
        ----------
        bins : int
            Number of bins to add
        '''
        self.bins += bins
        self.PSD = np.append(self.PSD, np.zeros(bins))

        self.max += bins * (self.PSDbounds[1] - self.PSDbounds[0])
        self.PSDbounds = np.linspace(self.min, self.max, self.bins+1)
        self.PSDsize = 0.5 * (self.PSDbounds[:-1] + self.PSDbounds[1:])

    def adjustSizeClassesEuler(self, checkDissolution = False):
        '''
        1) adds some bins to the end of the PSD if the last bin has at least 1 precipitate
            Number of bins is 1/4 of the original number of bins
        2) If adaptive bin size is enabled, then two checks
            2a) if number of bins > max bins, then resize to have the number of bins be the minimum
            2b) if checking dissolution and number of filled bins < 1/2 min bins,
                then resize to last filled bin with the number of bins being the maximum

        Parameters
        ----------
        checkDissolution : bool
            Whether to check if the PSD is getting smaller and resize accordingly

        Returns
        -------
        change : bool
            When the number of bins changed
        newIndices : int or None
            The number of bins added to the PSD
            If the size of the bins changed, then this is None to indicate that resizing occured
        '''
        change = False
        newIndices = None
        if self.PSD[-1] > 1:
            #print('adding bins')
            newIndices = self.bins
            self.addSizeClasses(int(self.originalBins/4))
            change = True

        if self._adaptiveBinSize:
            if self.bins > self.maxBins:
                #print('reducing bins')
                self.changeSizeClasses(self.PSDbounds[0], self.PSDbounds[-1], self.minBins)
                change = True
                newIndices = None
            elif checkDissolution and self.PSDbounds[-1] > 10*self.PSDbounds[0]:
                if any(self.PSD > 1) and np.amax(self.PSDsize[self.PSD > 1]) < self.PSDsize[int(self.minBins/2)]:
                    #print('splitting bins')
                    self.changeSizeClasses(self.PSDbounds[0], np.amax(self.PSDbounds[1:][self.PSD > 1]), self.maxBins)
                    change = True
                    newIndices = None
        return change, newIndices
        
    def Normalize(self):
        '''
        Normalizes the PSD to 1
        '''
        total = self.WeightedMoment(0, self.PSDbounds[1:] - self.PSDbounds[:-1])
        self.PSD /= total
            
    def NormalizeToMoment(self, order = 0):
        '''
        Normalizes the PSD with so that the moment of specified order will be 1
        
        Parameters
        ----------
        order : int (optional)
            Order of moment that PSD will be normalized to (defaults to 0)
            Using zeroth order will normalize the PSD so the sum of PSD will be 1
        '''
        total = self.Moment(order)
        self.PSD /= total

    def getDissolutionIndex(self, maxDissolution, minIndex = 0):
        '''
        Finds indices when the volume fraction of particles below this index is 
        within the maximum amount (fraction-wise) that the PSD is allowed to dissolve

        So find R_max where int(0, R_max, R^3 * dr) < maxDissolution * int(0, infinity, R^3 * dr)
        The index is the correspoinding index to R_max

        Parameters
        ----------
        maxDissolution : float
            Max fraction allowed to dissolve
        minIndex : int
            Minimum index which below, all particles are allowed to dissolve
            Upper limit on dissolution index

        Returns
        -------
        max of [dissolution index, minIndex]
        '''
        dissFrac = maxDissolution * self.ThirdMoment()
        dissIndex = np.argmax(self.CumulativeMoment(3) > dissFrac) - 1
        if dissIndex < 0:
            dissIndex = 0
        return np.amax([np.argmax(self.CumulativeMoment(3) > dissFrac), minIndex])
        

    def getDTEuler(self, currDT, growth, dissolutionIndex, maxBinRatio = 0.4):
        '''
        Calculates time interval for Euler implementation
            dt < dR / (2 * growth rate)
        This ensures that at most, only half of particles in one size class can go to another

        Also finds dt such that the max delta in growth rate is 0.4 dR
            We could use 0.5 dR which is the upper limit
                (for a given bin, the max change in density would remove all particles, with 0.5 getting smaller and 0.5 getting bigger)
            But 0.4 dR should be slightly more stable

        TODO: allow variable ratio - this will make it more flexible for testing different time step constraints

        Parameters
        ----------
        currDT : float
            Current time interval, will be returned if it's smaller than what's given by the contraint
        growth : array of floats
            Growth rate, must have lenth of bins+1 (or length of PSDbounds)
        maxDissolution : float
            Maximum volume allowed to dissolve
        startIndex : int
            First index to look at for growth rate, all indices below startIndex will be ignored
        maxBinRatio : float (optional)
            Max ratio of particles in bin allowed to move to a nearby bin
            Default is 0.4
        '''
        self.maxRatio = maxBinRatio
        growthFilter = growth[dissolutionIndex:-1][self.PSD[dissolutionIndex:] > 0]

        if len(growthFilter) == 0:
            return currDT
        else:
            if np.amax(np.abs(growthFilter)) == 0:
                return currDT
            else:
                return self.maxRatio * (self.PSDbounds[1] - self.PSDbounds[0]) / np.amax(np.abs(growthFilter))
    
    def getdXdtEuler(self, flux, nucRate, nucRadius, psd):
        '''
        dn_i/dt = d(G*n)/dr + nucRate

        d(G*n)/dr is calculated from two conditions
        For positive growth rates - d(G*n)/dr|_i = n_i * flux_i / dr
        For negative growth rates - d(G*n)/dr|_i = n_(i-1) * flux_i / dr
        TODO : check that the two equations above represent the implementation

        Parameters
        ----------
        flux : numpy array (bins+1)
            Growth rate of particles in m/s
        nucRate : float
            Nucleation rate in #/m^3/s
        nucRadius : float
            Nucleation radius in m
        psd : numpy array (bins)
            Particle size distribution with number density #/m3

        Returns
        -------
        dXdt (bins) - corresponds to dn_i/dt
        '''
        self._netFlux = np.zeros(self.bins+1)
        fluxSign = np.sign(flux)
        fluxSign[fluxSign == -1] = 0
        dR = self.PSDbounds[1:] - self.PSDbounds[:-1]
        self._netFlux[:-1] += flux[:-1] * psd * (1-fluxSign[:-1]) / dR
        self._netFlux[1:] += flux[1:] * psd * fluxSign[1:] / dR

        dXdt = (self._netFlux[:-1] - self._netFlux[1:])

        #Find size class for nucleated particles
        nRad = np.argmax(self.PSDbounds > nucRadius) - 1
        dXdt[nRad] += nucRate

        return dXdt
    
    def correctdXdtEuler(self, dt, flux, nucRate, nucRadius, psd):
        '''
        Given dt, correct the net flux so PSD will not be negative
            Essentially, the total number of particles leaving a bin should be less than or equal to the number of particles in the bin

        Size of fluxes is bins+1 while for PSD, it is bins

        For fluxes on the right (positive) side of the PSD
            We limit fluxes so that J_i+1 * dt < PSD_i

        For fluxes on the left (negative) side of the PSD
            We limit fluxes so that -J_i * dt < PSD_i

        Normally, this wouldn't be an issue since the time step from getDtEuler would limit J_i*dt to less than half the number of particles in a bin
            But because we set a dissolution threshold to ignore (to prevent extremely small dt since growth rate scales by 1/r), the bins below
            the dissolution threshold will not follow the constraint we apply in getDtEuler

        Parameters
        ----------
        dt : float
            time step
        flux : numpy array (bins+1)
            Growth rate of particles in m/s
        nucRate : float
            Nucleation rate in #/m^3/s
        nucRadius : float
            Nucleation radius in m
        psd : numpy array (bins)
            Particle size distribution with number density #/m3

        Returns
        -------
        dXdt (bins) - corresponds to dn_i/dt corrected to avoid negative bins
        '''
        #indBelow = self._netFlux[1:-1]*dt < -psd[1:]
        #self._netFlux[1:-1][indBelow] = -psd[1:][indBelow] / dt
        #indAbove = self._netFlux[1:-1]*dt > psd[:-1]
        #self._netFlux[1:-1][indAbove] = psd[:-1][indAbove] / dt

        indBelow = self._netFlux[:-1]*dt < -psd
        self._netFlux[:-1][indBelow] = -psd[indBelow] / dt
        indAbove = self._netFlux[1:]*dt > psd
        self._netFlux[1:][indAbove] = psd[indAbove] / dt

        dXdt = (self._netFlux[:-1] - self._netFlux[1:])

        #Find size class for nucleated particles
        nRad = np.argmax(self.PSDbounds > nucRadius) - 1
        dXdt[nRad] += nucRate

        return dXdt
    
    def UpdatePBMEuler(self, time, newN):
        '''
        Updates PBM with new values

        Parameters
        ----------
        time : float
            New time
        newN : numpy array
            New number density
        '''
        self.PSD = newN
        self.PSD[self.PSD < 1] = 0
        self.record(time)

    def MomentFromN(self, N, order):
        '''
        Given arbtrary PSD, return moment

        Parameters
        ----------
        N : numpy array
            PSD / number density
        order : float
            Moment order
        '''
        return np.sum(N * self.PSDsize**order)
    
    def CumulativeMomentFromN(self, N, order):
        '''
        Given arbtrary PSD, return cumulative moment (from 0 to max)

        Parameters
        ----------
        N : numpy array
            PSD / number density
        order : float
            Moment order
        '''
        return np.cumsum(N * self.PSDsize**order)
    
    def WeightedMomentFromN(self, N, order, weights):
        '''
        Given arbtrary PSD, return weighted moment

        Parameters
        ----------
        N : numpy array
            PSD / number density
        order : float
            Moment order
        weights : numpy array
            Weights for each bin
        '''
        return np.sum(N * self.PSDsize**order * weights)
    
    def CumulativeWeightedMomentFromN(self, N, order, weights):
        '''
        Given arbtrary PSD, return cumulative weighted moment (from 0 to max)

        Parameters
        ----------
        N : numpy array
            PSD / number density
        order : float
            Moment order
        weights : numpy array
            Weights for each bin
        '''
        return np.cumsum(self.PSD * self.PSDsize**order * weights)
    
    def ZeroMomentFromN(self, N):
        '''
        Sum of N
        '''
        return self.MomentFromN(N, 0)
    
    def FirstMomentFromN(self, N):
        '''
        Length weighted moment of N
        '''
        return self.MomentFromN(N, 1)
        
    def SecondMomentFromN(self, N):
        '''
        Area weighted moment of N
        '''
        return self.MomentFromN(N, 2)
        
    def ThirdMomentFromN(self, N):
        '''
        Volume weighted moment of N
        '''
        return self.MomentFromN(N, 3)
        
    def Moment(self, order):
        '''
        Moment of specified order

        Parameters
        ----------
        order : int
            Order of moment
        '''
        return self.MomentFromN(self.PSD, order)

    def CumulativeMoment(self, order):
        '''
        Cumulative distribution using moment of specified order

        Parameters
        ----------
        order : int
            Order of moment
        '''
        return self.CumulativeMomentFromN(self.PSD, order)
        
    def WeightedMoment(self, order, weights):
        '''
        Weighted moment of specified order

        Parameters
        ----------
        order : int
            Order of moment
        weights : array
            Weights to apply to each size class
            Array size of (bins)
        '''
        return self.WeightedMomentFromN(self.PSD, order, weights)

    def CumulativeWeightedMoment(self, order, weights):
        '''
        Weighted moment of specified order

        Parameters
        ----------
        order : int
            Order of moment
        weights : array
            Weights to apply to each size class
            Array size of (bins)
        '''
        return self.CumulativeWeightedMomentFromN(self.PSD, order, weights)

    def ZeroMoment(self):
        '''
        Sum of the PSD
        '''
        return self.Moment(0)
    
    def FirstMoment(self):
        '''
        Length weighted moment
        '''
        return self.Moment(1)
        
    def SecondMoment(self):
        '''
        Area weighted moment
        '''
        return self.Moment(2)
        
    def ThirdMoment(self):
        '''
        Volume weighted moment
        '''
        return self.Moment(3)
        
    def PlotCurve(self, axes, fill = False, logX = False, logY = False, scale = 1, *args, **kwargs):
        '''
        Plots the PSD as a curve
        
        Parameters
        ----------
        axes : Axes
            Axis to plot on
        fill : bool (optional)
            Will fill area between PSD curve and x-axis (defaults to False)
        logX : bool (optional)
            Whether to set x-axis on log scale (defaults to False)
        logY : bool (optional)
            Whether to set y-axis on log scale (defaults to False)
        scale : float (optional)
            Scale factor for x-axis (defaults to 1)
            Note: this is for grain boundary nucleation where the
                reported precipitate radius differs from the radius
                determined by precipitate curvature
        *args, **kwargs - extra arguments for plotting
        '''
        if hasattr(scale, '__len__'):
            scale = np.interp(self.PSDsize, self.PSDbounds, scale)
        else:
            scale = scale * np.ones(len(self.PSDsize))

        if fill:
            axes.fill_between(self.PSDsize * scale, self.PSD, np.zeros(len(self.PSD)), *args, **kwargs)
        else:
            axes.plot(self.PSDsize * scale, self.PSD, *args, **kwargs)
        self.setAxes(axes, scale, logX, logY) 

    def PlotDistributionDensity(self, axes, fill = False, logX = False, logY = False, scale = 1, *args, **kwargs):
        '''
        Plots the distribution density as a curve
        Defined as N_i / (R_i+1/2 - R_i-1/2)
        
        Parameters
        ----------
        axes : Axes
            Axis to plot on
        fill : bool (optional)
            Will fill area between PSD curve and x-axis (defaults to False)
        logX : bool (optional)
            Whether to set x-axis on log scale (defaults to False)
        logY : bool (optional)
            Whether to set y-axis on log scale (defaults to False)
        scale : float (optional)
            Scale factor for x-axis (defaults to 1)
            Note: this is for grain boundary nucleation where the
                reported precipitate radius differs from the radius
                determined by precipitate curvature
        *args, **kwargs - extra arguments for plotting
        '''
        if hasattr(scale, '__len__'):
            scale = np.interp(self.PSDsize, self.PSDbounds, scale)
        else:
            scale = scale * np.ones(len(self.PSDsize))

        if fill:
            axes.fill_between(self.PSDsize * scale, self.PSD, np.zeros(len(self.PSD)), *args, **kwargs)
        else:
            axes.plot(self.PSDsize * scale, self.PSD / (self.PSDbounds[1:] - self.PSDbounds[:-1]), *args, **kwargs)
        
        #Set x-limits
        if logX:
            if self.min == 0:
                axes.set_xlim([self.PSDbounds[1]*scale[0], self.max*scale[-1]])
            else:
                axes.set_xlim([self.min*scale[0], self.max*scale[-1]])
            axes.set_xscale('log')
        else:
            axes.set_xlim([self.min*scale[0], self.max*scale[-1]]) 

        #Set y-limits
        if logY:
            axes.set_ylim([1e-1, np.amax([1.1 * np.amax(self.PSD / (self.PSDbounds[1:] - self.PSDbounds[:-1])), 1])])
            axes.set_yscale('log')
        else:
            axes.set_ylim([0, np.amax([1.1 * np.amax(self.PSD / (self.PSDbounds[1:] - self.PSDbounds[:-1])), 1])])

    def PlotKDE(self, axes, bw_method = None, fill = False, logX = False, logY = False, scale = 1, *args, **kwargs):
        '''
        Plots the kernel density estimation (KDE)
        
        Parameters
        ----------
        axes : Axes
            Axis to plot on
        bw_method : str (optional)
            Method to estimate bandwidth ('scott', 'silverman' or a scalar)
            Defaults to scipy's default for KDE
        fill : bool (optional)
            Will fill area between PSD curve and x-axis (defaults to False)
        logX : bool (optional)
            Whether to set x-axis on log scale (defaults to False)
        logY : bool (optional)
            Whether to set y-axis on log scale (defaults to False)
        scale : float (optional)
            Scale factor for x-axis (defaults to 1)
            Note: this is for grain boundary nucleation where the
                reported precipitate radius differs from the radius
                determined by precipitate curvature
        *args, **kwargs - extra arguments for plotting
        '''
        x = np.linspace(self.min, self.max, 1000) 
        if np.all(self.PSD == 0):
            y = np.zeros(x.shape)
        else:
            kernel = sts.gaussian_kde(self.PSDsize, bw_method = bw_method, weights = self.PSD)  
            y = kernel(x) * self.ZeroMoment() * (self.PSDbounds[1] - self.PSDbounds[0])

        if hasattr(scale, '__len__'):
            scale = np.interp(x, self.PSDbounds, scale)
        else:
            scale = scale * np.ones(len(x))
        
        if fill:
            axes.fill_between(x * scale, y, np.zeros(len(y)), *args, **kwargs)
        else:
            axes.plot(x * scale, y, *args, **kwargs)
        self.setAxes(axes, scale, logX, logY) 
            
    def PlotHistogram(self, axes, outline = 'outline bins', fill = True, logX = False, logY = False, scale = 1, *args, **kwargs):
        '''
        Plots the PSD as a histogram
        
        Parameters
        ----------
        axes : Axes
            Axis to plot on
        outline : str (optional)
            How to outline the bins ('no outline', 'outline bins', 'outline top')
            Defaults to 'outline bins'
        fill : bool (optional)
            Will fill area between PSD curve and x-axis (defaults to False)
        logX : bool (optional)
            Whether to set x-axis on log scale (defaults to False)
        logY : bool (optional)
            Whether to set y-axis on log scale (defaults to False)
        scale : float (optional)
            Scale factor for x-axis (defaults to 1)
            Note: this is for grain boundary nucleation where the
                reported precipitate radius differs from the radius
                determined by precipitate curvature
        *args, **kwargs - extra arguments for plotting
        '''
        if outline == 'outline bins':
            xCoord, yCoord = np.zeros(1 + 3 * self.bins), np.zeros(1 + 3 * self.bins)
            xCoord[0], xCoord[1::3], xCoord[2::3], xCoord[3::3] = self.PSDbounds[0], self.PSDbounds[:-1], self.PSDbounds[1:], self.PSDbounds[1:]
            yCoord[1::3], yCoord[2::3] = self.PSD, self.PSD
        else:
            xCoord, yCoord = np.zeros(1 + 2 * self.bins), np.zeros(1 + 2 * self.bins)
            xCoord[0], xCoord[1::2], xCoord[2::2] = self.PSDbounds[0], self.PSDbounds[:-1], self.PSDbounds[1:]
            yCoord[1::2], yCoord[2::2] = self.PSD, self.PSD

        if hasattr(scale, '__len__'):
            scale = np.interp(xCoord, self.PSDbounds, scale)
        else:
            scale = scale * np.ones(len(xCoord))

        if outline != 'no outline':
            axes.plot(xCoord * scale, yCoord, *args, **kwargs)
            if fill:
                axes.fill_between(xCoord * scale, yCoord, np.zeros(len(yCoord)), alpha=0.3, *args, **kwargs)
        else:
            axes.fill_between(xCoord * scale, yCoord, np.zeros(len(yCoord)), *args, **kwargs)
        self.setAxes(axes, scale, logX, logY)

    def PlotCDF(self, axes, logX = False, scale = 1, order = 0, *args, **kwargs):
        '''
        Plots cumulative size distribution
        
        Parameters
        ----------
        axes : Axes
            Axis to plot on
        logX : bool (optional)
            Whether to set x-axis on log scale (defaults to False)
        scale : float (optional)
            Scale factor for x-axis (defaults to 1)
            Note: this is for grain boundary nucleation where the
                reported precipitate radius differs from the radius
                determined by precipitate curvature
        order : int (optional)
            Moment of specified order
        *args, **kwargs - extra arguments for plotting
        '''
        if hasattr(scale, '__len__'):
            scale = np.interp(self.PSDsize, self.PSDbounds, scale)
        else:
            scale = scale * np.ones(len(self.PSDsize))

        axes.plot(self.PSDsize * scale, self.CumulativeMoment(order) / self.Moment(order), *args, **kwargs)
        self.setAxes(axes, scale, logX, False) 
        axes.set_ylim([0, 1])
        
    def setAxes(self, axes, scale = 1, logX = False, logY = False): 
        '''
        Sets x- and y-axis to linear or log scale
        
        Parameters
        ----------
        axes : Axis
            Axis to plot on
        logX : bool (optional)
            Whether to set x-axis on log scale (defaults to False)
        logY : bool (optional)
            Whether to set y-axis on log scale (defaults to False)
        '''    
        if logX:
            if self.min == 0:
                axes.set_xlim([self.PSDbounds[1]*scale[0], self.max*scale[-1]])
            else:
                axes.set_xlim([self.min*scale[0], self.max*scale[-1]])
            axes.set_xscale('log')
        else:
            axes.set_xlim([self.min*scale[0], self.max*scale[-1]])

        #Don't set y limits if the PSD is empty
        if any(self.PSD > 0): 
            if logY:
                axes.set_ylim([1e-1, np.amax([1.1 * np.amax(self.PSD), 1])])
                axes.set_yscale('log')
            else:
                axes.set_ylim([0, np.amax([1.1 * np.amax(self.PSD), 1])])
        
        axes.set_ylabel('Frequency')
            
                    
        
        


import numpy as np
from kawin.precipitation.Plot import getTimeAxis

class StrengthModel:
    '''
    Defines strength model

    Following implementation described in
    M.R. Ahmadi, E. Povoden-Karadeni, K.I. Oksuz, A. Falahati and E. Kozeschnik
        Computational Materials Science 91 (2014) 173-186

    6 contributions are accounted for
    For dislocation cutting, contributions are coherency, modulus, anti-phase boundary, stacking fault energy and interfacial energy
    For dislocation bowing, contribution is orowan

    Contributions can be added for all phases or for a single phase
    '''
    def __init__(self):
        self.rssName = 'rss'
        self.LsName = 'Ls'

        #Solid solution strengthening parameters
        self.ssweights = {}
        self.ssexp = 1

        #Base strength
        self.sigma0 = 0

        self.M = 2.24

        #Precipitate strength factors
        self.coherencyEffect = {'all': False}        #Coherency effect
        self.modulusEffect = {'all': False}          #Modulus effect
        self.APBEffect = {'all': False}              #Anti-phase boundary effect
        self.SFEffect = {'all': False}               #Stacking fault energy effect
        self.IFEffect = {'all': False}               #Interfacial energy effect
        self.orowanEffect = {'all': False}           #Non-shearable (Orowan) mechanism

        #Parameters for dislocation line tension, general shear strength and orowan strengthening
        #These parameters are for the matrix phase
        self.G, self.b, self.nu, self.ri, self.theta, self.psi = None, None, None, None, 90*np.pi/180, 120*np.pi/180

        #eps, GP, yAPB, ySFP and gamma are specific to a phase
        #Parameters for coherency effect
        self.eps = {}

        #Parameters for modulus effect
        self.Gp, self.w1, self.w2 = {}, 0.0722, 0.81

        #Parameters for anti-phase boundary effect
        self.yAPB, self.s, self.beta, self.V = {}, 2, 1, 2.8

        #Parameters for stacking fault energy effect
        self.ySFM, self.ySFP, self.bp = None, {}, {}

        #Parameters for interfacial effect
        self.gamma = {}

        #Model types for line tension and J
        self.T = self.Tcomplex
        self.J = self.Jsimple

        #Single phase superposition exponent
        self.singlePhaseExp = 1.8

        #Multi-phase superposition exponents
        self.multiphaseSameExp = 1.8
        self.multiphaseMixedExp = 1.4

        #Superposition exponent for total strength
        self.totalStrengthExp = 1.8

        #Strength terms
        self.rss = None
        self.ls = None
        self.solidStrength = None

    def save(self, filename, compressed = True):
        '''
        Saves strength model data

        Note, this only saves solid solution strength, the rss and ls terms
        Parameters should be free so user can load model and evaluate different parameters
        '''
        if compressed:
            np.savez_compressed(filename, ssStrength=self.solidStrength, rss = self.rss, ls = self.ls)
        else:
            np.savez(filename, ssStrength=self.solidStrength, rss = self.rss, ls = self.ls)

    def load(self, filename):
        data = np.load(filename)
        self.solidStrength = data['ssStrength']
        self.rss = data['rss']
        self.ls = data['ls']

    def _getStrengthFunctions(self, selectedContributions = None):
        '''
        Internal function that creates arrays for dislocation cutting mechanisms

        wfuncs, sfuncs - list of functions for each contribution for weak and strong effects
        contributions - each contribution has a dictionary of str : boolean to say whether a phase has that contribution
        labels - labels for plotting

        Parameters
        ----------
        selectedContributions : None or List[str]
            If None, will return weak/strong functions and labels for all contributions
            If List[str], will return weak/strong functions and labels for only the contributions defined in list
                Options are: Coherency, Modulus, APB, SFE and/or Interfacial
        
        Returns
        -------
        wfuncs - List of functions for weak contributions
        sfuncs - List of functions for strong contributions
        contributions - List of {phase str:boolean} for whether the contribution is enabled
        labels - List of labels for plotting
        '''
        wfuncs = [self.coherencyWeak, self.modulusWeak, self.APBweak, self.SFEweak, self.interfacialWeak]
        sfuncs = [self.coherencyStrong, self.modulusStrong, self.APBstrong, self.SFEstrong, self.interfacialStrong]
        contributions = [self.coherencyEffect, self.modulusEffect, self.APBEffect, self.SFEffect, self.IFEffect]
        labels = ['Coherency', 'Modulus', 'APB', 'SFE', 'Interfacial']
        if selectedContributions is None:
            return wfuncs, sfuncs, contributions, labels
        else:
            wfuncsSub, sfuncsSub, contributionsSub, labelsSub = [], [], [], []
            lowerLabels = [l.lower() for l in labels]
            for c in selectedContributions:
                if c.lower() in lowerLabels:
                    index = lowerLabels.index(c.lower())
                    wfuncsSub.append(wfuncs[index])
                    sfuncsSub.append(sfuncs[index])
                    contributionsSub.append(contributions[index])
                    labelsSub.append(labels[index])
            return wfuncsSub, sfuncsSub, contributionsSub, labelsSub 
        

    def setBaseStrength(self, sigma0):
        '''
        Sets base strength of matrix

        Parameters
        ----------
        sigma0 : float
            Base strength (Pa)
        '''
        self.sigma0 = sigma0

    def setSolidSolutionStrength(self, weights, exp = 1):
        '''
        Parameters for solid solution strengthening

        Parameters
        ----------
        weights : dictionary
            Dictionary mapping element (str) to weight (float). Weights are in units of (Pa)
        exp : float
            Exponential factor for weights
        '''
        self.ssweights = weights
        self.ssexp = exp

    def setTaylorFactor(self, M = 2.24):
        '''
        Taylor factor for converting critical resolved shear stress to yield strength

        Default is 2.24
        '''
        self.M = M

    def setStrengthSuperpositionExponent(self, singlePhaseExp = 1.8, multiPhaseSameExp = 1.8, multiPhaseMixedExp = 1.4, totalExp = 1.8):
        '''
        Sets exponent (n) for superposition function when combining strength contributions

        sigma^n = sigma_1^n + sigma_2^n + sigma_3^n + ...

        Parameters
        ----------
        singlePhaseExp : float (optional)
            Exponent for adding strength contributions for a single precipitate phase
            Default = 1.8
        multiPhaseSameExp : float (optional)
            Exponent for adding strength contributions for multiple precipitate phases when the 
            strengthening mechanisms are the same
            Default = 1.8
        multiPhaseMixedExp : float (optional)
            Exponent for adding strength contributions for multiple precipitate phases when the
            strengthening mechanisms differ
            Default = 1.4
        totalExp : float (optional)
            Exponent for adding strength mechanisms for total strength
            This includes base strength, solid solution strengthening and precipitation hardening
            Default = 1.8
        '''
        self.singlePhaseExp = singlePhaseExp
        self.multiphaseSameExp = multiPhaseSameExp
        self.multiphaseMixedExp = multiPhaseMixedExp
        self.totalStrengthExp = totalExp

    def setDislocationParameters(self, G, b, nu = 1/3, ri = None, theta = 90, psi = 120):
        '''
        Parameters for dislocation line tension, used for all precipitate strength models

        Parameters
        ----------
        G : float
            Shear modulus of matrix (Pa)
        b : float
            Burgers vector (meters)
        nu : float
            Poisson ratio
        ri : float (optional)
            Dislocation core radius (meters)
            If None, ri will be set to Burgers vector
        r0 : float (optional)
            Closest distance between parallel dislocations
            For shearable precipitates, r0 is average distance between particles on slip plane
            For non-shearable precipitates, r0 is average particle diameter on slip plane
            If None, r0 will be set such that ln(r0/ri) = 2*pi
        theta : float (optional)
            Dislocation characteristic, 90 for edge, 0 for screw (default is 90)
        psi : float (optional)
            Dislocation bending angle (default is 120)
        '''
        self.G = G
        self.b = b
        self.nu = nu
        self.ri = b if ri is None else ri
        self.theta = theta * np.pi/180
        self.psi = psi * np.pi/180

    def setCoherencyParameters(self, eps, phase = 'all'):
        '''
        Parameters for coherency effect

        Parameters
        ----------
        eps : float
            Lattice misfit strain
        phase : str (optional)
            Defaults to 'all'
            If 'all', contribution and parameters will be applied to all phases
            If name of a specific phase (must be one that is defined in the PrecipitateModel)
                contribution and parameters will only be applied to that phase
        '''
        self.coherencyEffect[phase] = True
        self.eps[phase] = eps

    def setModulusParameters(self, Gp, w1 = 0.05, w2 = 0.85, phase = 'all'):
        '''
        Parameters for modulus effect

        Parameters
        ----------
        Gp : float
            Shear modulus of precipitate
        w1 : float (optional)
            First factor for Nembach model taking value between 0.0175 and 0.0722
            Default at 0.05
        w2 : float (optional)
            Second factor for Nembach model taking value of 0.81 +/- 0.09
            Default at 0.85
        phase : str (optional)
            Defaults to 'all'
            If 'all', contribution and parameters will be applied to all phases
            If name of a specific phase (must be one that is defined in the PrecipitateModel)
                contribution and parameters will only be applied to that phase
        '''
        self.modulusEffect[phase] = True
        self.Gp[phase] = Gp
        self.w1 = w1
        self.w2 = w2

    def setAPBParameters(self, yAPB, s = 2, beta = 1, V = 2.8, phase = 'all'):
        '''
        Parameters for anti-phase boundary effect for ordered precipitates in a disordered matrix

        Parameters
        ----------
        yAPB : float
            Anti-phase boundary energy
        s : int (optional)
            Number of leading + trailing dislocations to repair anti-phase boundary
            Default at 2
        beta : float (optional)
            Factor representating dislocation shape (between 0 and 1)
            Default at 1
        V : float (optional)
            Correction factor accounting for extra dislocations and uncertainties
            Default at 2.8
        phase : str (optional)
            Defaults to 'all'
            If 'all', contribution and parameters will be applied to all phases
            If name of a specific phase (must be one that is defined in the PrecipitateModel)
                contribution and parameters will only be applied to that phase
        '''
        self.APBEffect[phase] = True
        self.yAPB[phase] = yAPB
        self.s = s
        self.beta = beta
        self.V = V

    def setSFEParameters(self, ySFM, ySFP, bp = None, phase = 'all'):
        '''
        Parameters for stacking fault energy effect

        Parameters
        ----------
        ySFM : float
            Stacking fault energy of matrix
        ySFP : float
            Stacking fault energy of precipitate
        bp : float (optional)
            Burgers vector in precipitate
            If None, will be set to burgers vector in matrix
        phase : str (optional)
            Defaults to 'all'
            If 'all', contribution and parameters will be applied to all phases
            If name of a specific phase (must be one that is defined in the PrecipitateModel)
                contribution and parameters will only be applied to that phase
        '''
        self.SFEffect[phase] = True
        self.ySFM = ySFM
        self.ySFP[phase] = ySFP
        self.bp[phase] = self.b if bp is None else bp

    def setInterfacialParameters(self, gamma, phase = 'all'):
        '''
        Parameters for interfacial effect

        Parameters
        ----------
        gamma : float
            Interfacial energy of matrix/precipitate surface
        phase : str (optional)
            Defaults to 'all'
            If 'all', contribution and parameters will be applied to all phases
            If name of a specific phase (must be one that is defined in the PrecipitateModel)
                contribution and parameters will only be applied to that phase
        '''
        self.IFEffect[phase] = True
        self.gamma[phase] = gamma

    def setTmodel(self, modelType = 'complex'):
        '''
        Set model for line tension

        Parameters
        ----------
        modelType : str
            'complex' for line tension based off dislocation character
            'simple' for simple line tension model (T = G * b^2 / 2)
        '''
        if modelType == 'simple':
            self.T = self.Tsimple
        else:
            self.T = self.Tcomplex

    def setJfactor(self, modelType = 'complex'):
        '''
        Set model for J

        Parameters
        ----------
        modelType : str
            'complex' for J based off dislocation character
            'simple' for J = 1
        '''
        if modelType == 'simple':
            self.J = self.Jsimple
        else:
            self.J = self.Jcomplex

    def epsMisfit(self, delta, phase='all'):
        '''
        Strain from lattice misfit
        '''
        self.eps['all'] = (1/3) * (1+self.nu) / (1-self.nu) * delta

    def Tcomplex(self, theta, r0):
        '''
        Dislocation line tension
        '''
        return self.G*self.b**2 / (4*np.pi) * (1 + self.nu - 3*self.nu*np.sin(theta)**2) / (1 - self.nu) * np.log(r0 / self.ri)

    def Tsimple(self, theta, r0):
        '''
        Simple model for dislocation tension
        '''
        return 0.5 * self.G * self.b**2

    @property
    def Jcomplex(self):
        '''
        Correction constant for mean distance between randomly arranged particles accounting for dislocation characteristic
        '''
        return (1 - self.nu * np.cos(np.pi/2 - self.theta)**2) / np.sqrt(1 - self.nu)

    @property
    def Jsimple(self):
        '''
        Same as Jcomplex but assumes edge dislocation
        '''
        return 1

    def coherencyWeak(self, r, Ls, r0, phase='all'):
        '''
        Coherency effect for mixed dislocation on weak and shearable particles
        '''
        return (1.3416*np.cos(self.theta)**2 + 4.1127*np.sin(self.theta)**2) / Ls * np.sqrt(self.G**3 * self.eps[phase]**3 * r**3 * self.b / self.T(self.theta, r0))

    def coherencyStrong(self, r, Ls, r0, phase='all'):
        '''
        Coherency effect for mixed dislocation on strong and shearable particles
        '''
        return (2*np.cos(self.theta)**2 + 2.1352*np.sin(self.theta)**2) / Ls * np.power(self.T(self.theta, r0)**3 * self.G * self.eps[phase] * r / self.b**3, 1/4)

    def Fmod(self, r, phase='all'):
        '''
        Term for modulus effect
        '''
        return self.w1 * np.abs(self.G - self.Gp[phase]) * self.b**2 * np.power(r / self.b, self.w2)

    def modulusWeak(self, r, Ls, r0, phase='all'):
        '''
        Modulus effect for mixed dislocation on weak and shearable particles
        '''
        return 2 * self.T(self.theta, r0) / (self.b * Ls) * np.power(self.Fmod(r, phase) / (2 * self.T(self.theta, r0)), 3/2)

    def modulusStrong(self, r, Ls, r0, phase='all'):
        '''
        Modulus effect for edge or screw dislocation on strong and shearable particles
        '''
        return self.J * self.Fmod(r, phase) / (self.b * Ls)

    def APBweak(self, r, Ls, r0, phase='all'):
        '''
        Anti-phase boundary effect for mixed dislocation on weak and shearable particles
        '''
        return 2 / (self.s * self.b * Ls) * (2 * self.T(self.theta, r0) * np.power(r * self.yAPB[phase] / self.T(self.theta, r0), 3/2) - 16 * self.beta * self.yAPB[phase] * r**2 / (3 * np.pi * Ls))

    def APBstrong(self, r, Ls, r0, phase='all'):
        '''
        Anti-phase boundary effect for mixed dislocation on strong and shearable particles
        '''
        return 0.69 / (self.b * Ls) * np.sqrt(8 * self.V * self.T(self.theta, r0) * r * self.yAPB[phase] / 3)

    def K(self, theta, phase='all'):
        return self.G * self.bp[phase]**2 * (2 - self.nu - 2 * self.nu * np.cos(2 * theta)) / (8 * np.pi * (1 - self.nu))

    def SFEWeff(self, theta, phase='all'):
        '''
        Effective stacking fault width for mixed dislocations
        '''
        return 2* self.K(theta, phase) / (self.ySFM + self.ySFP[phase])

    def SFEFterm(self, r, phase='all'):
        '''
        Stacking fault term
        '''
        return 2 * (self.ySFM - self.ySFP[phase]) * np.sqrt(self.SFEWeff(self.theta, phase) * r - self.SFEWeff(self.theta, phase)**2 / 4)

    def interfacialWeak(self, r, Ls, r0, phase='all'):
        '''
        Interfacial energy effect for mixed dislocation on weak and shearable particles
        '''
        return 2 * self.T(self.theta, r0) / (self.b * Ls) * np.power(2 * self.gamma[phase] * self.b / (2 * self.T(self.theta, r0)), 3/2)

    def interfacialStrong(self, r, Ls, r0, phase='all'):
        '''
        Interfacial energy effect for mixed dislocations on strong and shearable particles
        '''
        return 2 * self.gamma[phase] / Ls

    def SFEweak(self, r, Ls, r0, phase='all'):
        '''
        Stacking fault energy effect for mixed dislocations on weak and shearable particles
        '''
        return 2 * self.T(self.theta, r0) / (self.b * Ls) * np.power(self.SFEFterm(r, phase) / (2 * self.T(self.theta, r0)), 3/2)

    def SFEstrong(self, r, Ls, r0, phase='all'):
        '''
        Stacking fault energy effect for mixed dislocations on strong and shearable particles
        '''
        return self.SFEFterm(r, phase) / (self.b * Ls)

    def orowan(self, r, Ls):
        '''
        Orowan strengthening for non-shearable particles
        '''
        return self.J * self.G * self.b / (2 * np.pi * np.sqrt(1 - self.nu) * Ls) * np.log(2 * r / self.ri)

    def ssStrength(self, model, n):
        '''
        Solid solution strength model
        sigma_ss = sum(k_i * c_i^n)

        Parameters
        ----------
        model : KWNEuler object
            Model to take composition from

        Returns
        -------
        strength : array of floats
            Solid solution strength contribution over time
        '''
        val = 0
        for i in range(len(model.elements)):
            if model.elements[i] in self.ssweights:
                val += self.ssweights[model.elements[i]]*model.xComp[n,i]**self.ssexp
        return val

    def rssterm(self, model, p):
        '''
        Mean projected radius of particles

        r1 = first ordered moment of particle size distribution
        r2 = second ordered moment of particle size distribution
        rss = sqrt(2/3) * r2 / r1

        Parameters
        ----------
        model : PrecipitateModel
        p : int
            Phase index
        i : int
            Iteration of the model
        '''
        r1 = np.sum(model.PBM[p].PSD * model.PBM[p].PSDsize)
        r2 = np.sum(model.PBM[p].PSD * model.PBM[p].PSDsize**2)
        if r1 == 0:
            rss = 0
        else:
            rss = np.sqrt(2/3) * r2 / r1
        return rss

    def Lsterm(self, model, p):
        '''
        Mean surface to surface distance between particles

        r1 = first ordered moment of particle size distribution
        r2 = second ordered moment of particle size distribution
        rss = sqrt(2/3) * r2 / r1
        ls = sqrt(ln(3)/(2*pi*r1) + (2*rss)^2) - 2*rss

        Parameters
        ----------
        model : PrecipitateModel
        p : int
            Phase index
        i : int
            Iteration of the model
        '''
        r1 = np.sum(model.PBM[p].PSD * model.PBM[p].PSDsize)
        r2 = np.sum(model.PBM[p].PSD * model.PBM[p].PSDsize**2)
        if r1 == 0:
            Ls = 0
        else:
            rss = np.sqrt(2/3) * r2 / r1
            Ls = np.sqrt(np.log(3) / (2*np.pi*r1) + (2*rss)**2) - 2*rss
        return Ls
    
    def updateCoupledModel(self, model):
        '''
        Computes rss, ls and solid solution strengthening terms
        from current state of the PrecipitateModel

        Parameters
        ----------
        model : PrecpitateModel
        '''
        if self.rss is None:
            self.rss = np.zeros((1, len(model.phases)))
            self.ls = np.zeros((1, len(model.phases)))
            self.solidStrength = np.zeros(1)
            self.solidStrength[0] = self.ssStrength(model, 0)

        self.rss = np.append(self.rss, [[self.rssterm(model, p) for p in range(len(model.phases))]], axis=0)
        self.ls = np.append(self.ls, [[self.Lsterm(model, p) for p in range(len(model.phases))]], axis=0)
        self.solidStrength = np.append(self.solidStrength, [self.ssStrength(model, model.n)], axis=0)

    def precStrength(self, model):
        '''
        Gets strength as a function of time

        Parameters
        ----------
        model : PrecipitateModel
        '''
        #rss = model.getAdditionalOutput(self.rssName)
        #Ls = model.getAdditionalOutput(self.LsName)
        rss = self.rss
        Ls = self.ls

        ps = []
        totalCompare = np.zeros(len(rss[:,0]))
        for i in range(len(model.phases)):
            weakContributions, strongContributions, orowan, _ = self.getStrengthContributions(rss[:,i], Ls[:,i], model.phases[i])
            strength, compare, _ = self.combineStrengthContributions(weakContributions, strongContributions, orowan, returnComparison=True)
            compare[~np.isfinite(strength)] = 0
            strength[~np.isfinite(strength)] = 0
            ps.append(strength)
            totalCompare += np.array(compare, dtype='int')
        ps = np.array(ps)
        totalStrength = np.zeros(len(ps[0]))
        indices = (totalCompare == 0) | (totalCompare == len(model.phases))
        totalStrength[indices] = np.power(np.sum(np.power(ps[:,indices], self.multiphaseSameExp), axis=0), 1/self.multiphaseSameExp)
        totalStrength[~indices] = np.power(np.sum(np.power(ps[:,~indices], self.multiphaseMixedExp), axis=0), 1/self.multiphaseMixedExp)
        return totalStrength

    def getStrengthContributions(self, rss, Ls, phase = 'all', selectedContributions=None):
        '''
        Gets strength contributions from a model

        Parameters
        ----------
        rss : array
            Mean projected radius
        Ls : array
            Mean surface to surface particle spacing
        phase : str (optional)
            Phase name
            Defaults to 'all'
        selectedContributions : None or List[str]
            If None, will return weak/strong functions and labels for all contributions
            If List[str], will return weak/strong functions and labels for only the contributions defined in list
                Options are: Coherency, Modulus, APB, SFE and/or Interfacial
        '''
        r0Weak = Ls / np.sqrt(np.cos(self.psi / 2))
        r0Strong = Ls
        weakContributions = []
        strongContributions = []
        contributionsList = []
        wfuncs, sfuncs, contributions, ylabel = self._getStrengthFunctions(selectedContributions)
        for i in range(len(wfuncs)):
            if contributions[i]['all'] or (phase in contributions[i] and contributions[i][phase]):
                with np.errstate(divide='ignore', invalid='ignore'):
                    if (phase in contributions[i] and contributions[i][phase]):
                        weakContributions.append(wfuncs[i](rss, Ls, r0Weak, phase))
                        strongContributions.append(sfuncs[i](rss, Ls, r0Strong, phase))
                    else:
                        weakContributions.append(wfuncs[i](rss, Ls, r0Weak, 'all'))
                        strongContributions.append(sfuncs[i](rss, Ls, r0Strong, 'all'))
                    contributionsList.append(ylabel[i])
        weakContributions = np.array(weakContributions)
        weakContributions[(weakContributions < 0) | ~np.isfinite(weakContributions)] = 0
        strongContributions = np.array(strongContributions)
        strongContributions[(strongContributions < 0) | ~np.isfinite(strongContributions)] = 0
        tauowo = np.array(self.orowan(rss, Ls))
        tauowo[~np.isfinite(tauowo)] = 0
        return weakContributions, strongContributions, tauowo, contributionsList
    
    def combineStrengthContributions(self, weakContributions, strongContributions, orowan, returnComparison = False):
        '''
        Combines weak, strong and orowan contributions

        Parameters
        ----------
        weakContributions : 2D array
            List of arrays for weak contribution mechanisms
        strongContributions : 2D array
            List of arrays for strong contribution mechanisms
        orowan : array
            Orowan strengthening contribution
        returnComparison : boolean (optional)
            If True, returns additional array of boolean where
                True - weak contribution is higher than strong and orowan contributions
                False - weak contribution is lower than strong and orowan contributions
        '''
        tausumweak = np.zeros(orowan.shape) if len(weakContributions) == 0 else np.array(np.power(np.sum(np.power(weakContributions, self.singlePhaseExp), axis=0), 1/self.singlePhaseExp))
        tausumstrong = np.zeros(orowan.shape) if len(strongContributions) == 0 else np.array(np.power(np.sum(np.power(strongContributions, self.singlePhaseExp), axis=0), 1/self.singlePhaseExp))
        tausumweak[~np.isfinite(tausumweak)] = 0
        tausumstrong[~np.isfinite(tausumstrong)] = 0
        orowan[~np.isfinite(orowan)] = 0
        taumin = np.amin(np.array([tausumweak, tausumstrong, orowan]), axis=0)
        if returnComparison:
            return self.M * taumin, (tausumweak > tausumstrong) & (tausumweak > orowan), (self.M * tausumweak, self.M * tausumstrong, self.M * orowan)
        else:
            return self.M * taumin

    def totalStrength(self, ssStrength, precStrength):
        '''
        Combined base strength, solid solution strength and precipitate strength

        Parameters
        ----------
        ssStrength : array
            Solid solution strengthening
        precStrength : array
            Precipitate strengthening
        '''
        sigma0 = self.sigma0*np.ones(len(ssStrength))
        return np.power(np.sum(np.power([sigma0, ssStrength, precStrength], self.totalStrengthExp), axis=0), 1/self.totalStrengthExp)

    def getStrengthUnits(self, strengthUnits = 'Pa'):
        '''
        Internal function to return scale and label for the y-axis based off units of strength
        '''
        yscale = 1
        ylabel = 'Strength (Pa)'
        if strengthUnits.lower() == 'kpa':
            yscale = 1e3
            ylabel = 'Strength (kPa)'
        elif strengthUnits.lower() == 'mpa':
            yscale = 1e6
            ylabel = 'Strength (MPa)'
        elif strengthUnits.lower() == 'gpa':
            yscale = 1e9
            ylabel = 'Strength (GPa)'
        return yscale, ylabel

    def plotPrecipitateStrengthOverR(self, ax, r, Ls, phase=None, strengthUnits = 'MPa', contribution = None, *args, **kwargs):
        '''
        Plots precipitate strength contribution as a function of radius

        Parameters
        ----------
        ax : Axis
        r : list
            Equivalent radius
        Ls : list
            Surface to surface particle distance
        strengthUnits : str
            Units for strength, options are 'Pa', 'kPa', 'MPa' or 'GPa'
        contribution : None or str
            If None, will plot overall strength
            If str, will plot selected contribution or all contributions
                Options are: Coherency, Modulus, APB, SFE or Interfacial
        '''
        if phase is None:
            phase = 'all'

        self.plotPrecipitateStrengthOverX(ax, r, r, Ls, phase, strengthUnits, contribution, *args, **kwargs)
        ax.set_xlabel('Radius (m)')

    def plotPrecipitateStrengthOverTime(self, ax, model, phase = None, bounds = None, timeUnits = 's', strengthUnits = 'MPa', contribution = None, *args, **kwargs):
        '''
        Plots precipitate strength contribution as a function of time

        Parameters
        ----------
        ax : Axis
        r : list
            Equivalent radius
        Ls : list
            Surface to surface particle distance
        strengthUnits : str
            Units for strength, options are 'Pa', 'kPa', 'MPa' or 'GPa'
        contribution : None or str
            If None, will plot overall strength
            If str, will plot selected contribution or all contributions
                Options are: Coherency, Modulus, APB, SFE or Interfacial
        '''
        timeScale, timeLabel, bounds = getTimeAxis(model, timeUnits, bounds)

        self.plotPrecipitateStrengthOverX(ax, model.time*timeScale, self.rss, self.ls, phase, strengthUnits, contribution, *args, **kwargs)
        ax.set_xlabel(timeLabel)
        ax.set_xscale('log')
        ax.set_xlim(bounds)

    def plotPrecipitateStrengthOverX(self, ax, x, r, Ls, phase = None, strengthUnits = 'MPa', contribution = None, *args, **kwargs):
        '''
        Plots precipitate strength contribution as a function of x

        TODO: make this a bit more generalized where you can set the contribution you want to plot
            This should also remove the restriction that axes subplot must be 3x2

        Parameters
        ----------
        ax : Axis
        x : list
            X coordinates to plot against
        r : list
            Equivalent radius, must correspond to x
        Ls : list
            Surface to surface particle distance, must correspond to x
        strengthUnits : str
            Units for strength, options are 'Pa', 'kPa', 'MPa' or 'GPa'
        contribution : None or str
            If None, will plot overall strength
            If str, will plot selected contribution or all contributions
                Options are: Coherency, Modulus, APB, SFE, Interfacial, Orowan or All
        '''
        yscale, ylabel = self.getStrengthUnits(strengthUnits)
        if contribution is not None:
            if contribution.lower() == 'orowan':
                tauowo = np.array(self.orowan(r, Ls))
                tauowo[~np.isfinite(tauowo)] = 0
                ax.plot(x, self.M * tauowo / yscale, *args, **kwargs)
                ax.set_ylabel(r'$\tau_{orowan}$ (' + strengthUnits + ')')
                ax.set_ylim(bottom=0)
                ax.set_xlim([x[0], x[-1]])

            elif contribution.lower() != 'all':
                _, _, _, ylabel = self._getStrengthFunctions([contribution])
                weak, strong, oro, contributionList = self.getStrengthContributions(r, Ls, phase, [contribution])
                if ylabel[0] in contributionList:
                    ax.plot(x, self.M * weak[0] / yscale, x, self.M * strong[0] / yscale, *args, **kwargs)
                    ax.set_ylim(bottom=0)
                    ax.legend(['Weak', 'Strong'])
                else:
                    ax.plot(x, np.zeros(len(x)), *args, **kwargs)
                    ax.set_ylim([-1, 1])
                ax.set_ylabel(r'$\tau_{' + ylabel[0] + '}$ (' + strengthUnits + ')')
                ax.set_xlim([x[0], x[-1]])

            else:
                _, _, _, ylabel = self._getStrengthFunctions()
                weak, strong, oro, contributionList = self.getStrengthContributions(r, Ls, phase)
                strength, _, summedContributions = self.combineStrengthContributions(weak, strong, oro, returnComparison=True)
                wtot, stot, oro = summedContributions

                ax.plot(x, wtot/yscale, x, stot/yscale, x, oro/yscale, x, strength/yscale, *args, **kwargs)
                ax.set_ylim(bottom=0)
                ax.set_ylabel(r'$\tau$ (' + strengthUnits + ')')
                ax.legend(['Weak', 'Strong', 'Orowan', 'Minimum'])
                ax.set_xlim([x[0], x[-1]])
            

        else:
            weak, strong, oro, contributionList = self.getStrengthContributions(r, Ls, phase)
            strength = self.combineStrengthContributions(weak, strong, oro)
            ax.plot(x, strength / yscale, *args, **kwargs)
            ax.set_ylabel('Yield ' + ylabel)
            ax.set_ylim(bottom=0)
            ax.set_xlim([x[0], x[-1]])

    def plotStrength(self, ax, model, plotContributions = False, bounds = None, timeUnits = 's', strengthUnits = 'MPa', *args, **kwargs):
        '''
        Plots strength over time

        Parameters
        ----------
        ax : Axis
        model : KWNEuler object
        plotContributions : boolean
            Whether to plot each contribution of the strength model (defaults to False)
        bounds : tuple or None
            Bounds on time axis (if None, the bounds will automatically be set)
        timeUnits : str
            Units of time to plot in, options are 's' for seconds, 'm' for minutes or 'h' for hours
        strengthUnits : str
            Units for strength, options are 'Pa', 'kPa', 'MPa' or 'GPa'
        '''
        timeScale, timeLabel, bounds = getTimeAxis(model, timeUnits, bounds)
        yscale, ylabel = self.getStrengthUnits(strengthUnits)

        sigma0 = self.sigma0 * np.ones(len(model.time))
        #ssStrength = self.ssStrength(model) if len(self.ssweights) > 0 else np.zeros(len(model.time))
        ssStrength = self.solidStrength
        precStrength = self.precStrength(model)

        total = self.totalStrength(ssStrength, precStrength)

        ax.plot(model.time*timeScale, total / yscale, *args, **kwargs)

        if plotContributions:
            ax.plot(model.time*timeScale, sigma0 / yscale, *args, **kwargs)
            ax.plot(model.time*timeScale, ssStrength / yscale, *args, **kwargs)
            ax.plot(model.time*timeScale, precStrength / yscale, *args, **kwargs)
            ax.legend(['Total Strength', r'$\sigma_0$', 'SS Strength', 'Precipitate Strength'])

        ax.set_xlabel(timeLabel)
        ax.set_xlim(bounds)
        ax.set_ylabel(ylabel)
        ax.set_ylim(bottom=0)
        ax.set_xscale('log')


    '''
    Comparison functions

    Functions used for the strength model are for mixed dislocations from a MatCalc presentation

    The following functions are from Ahmadi et al and are for specific cases regarding dislocation behavior
    '''
    def coherencyWeakEdge(self, r, Ls, r0, phase='all'):
        '''
        Coherency effect for edge dislocation on weak and shearable particles
        '''
        return np.sqrt((592 / 35) * self.G**3 * self.b * self.eps[phase]**3 * r**3 / (Ls**2 * self.T(np.pi/2, r0)))

    def coherencyWeakScrew(self, r, Ls, r0, phase='all'):
        '''
        Coherency effect for screw dislocation on weak and shearable particles
        '''
        return np.sqrt((9/5) * self.G**3 * self.b * self.eps[phase]**3 * r**3 / (Ls**2 * self.T(0, r0)))

    def coherencyStrongEdge(self, r, Ls, r0, phase='all'):
        '''
        Coherency effect for edge dislocation on strong and shearable particles
        '''
        return np.sqrt(2) * np.power(3, 3/8) * self.J / Ls * np.power((self.T(np.pi/2, r0)**3 * self.G * self.eps[phase] * r) / self.b**3, 1/4)

    def coherencyStrongScrew(self, r, Ls, r0, phase='all'):
        '''
        Coherency effect for screw dislocation on strong and shearable particles
        '''
        return 2 * self.J / Ls * np.power((self.T(0, r0)**3 * self.G * self.eps[phase] * r) / self.b**3, 1/4)

    def modulusWeakEdge(self, r, Ls, r0, phase='all'):
        '''
        Modulus effect for edge dislocation on weak and shearable particles
        '''
        return 2 * self.T(np.pi/2, r0) / (self.b * Ls) * np.power(self.w1 * np.abs(self.Gp[phase] - self.G) * self.b**2 * np.power(r/self.b, self.w2) / (2*self.T(np.pi/2, r0)), 3/2)

    def modulusWeakScrew(self, r, Ls, r0, phase='all'):
        '''
        Modulus effect for screw dislocation on weak and shearable particles
        '''
        return 2 * self.T(0, r0) / (self.b * Ls) * np.power(self.w1 * np.abs(self.Gp[phase] - self.G) * self.b**2 * np.power(r/self.b, self.w2) / (2*self.T(0, r0)), 3/2)

    def APBweakEdge(self, r, Ls, r0, phase='all'):
        '''
        Anti-phase boundary effect for edge dislocation on weak and shearable particles
        '''
        xi = 16 * self.yAPB[phase] * r**2 / (3 * np.pi * self.b * Ls**2)
        return 2/self.s * (2 * self.T(np.pi/2, r0) / (self.b * Ls) * np.power(2 * self.yAPB[phase] * r / (2*self.T(np.pi/2, r0)), 3/2) - self.beta * xi)

    def APBweakScrew(self, r, Ls, r0, phase='all'):
        '''
        Anti-phase boundary effect for screw dislocation on weak and shearable particles
        '''
        xi = 16 * self.yAPB[phase] * r**2 / (3 * np.pi * self.b * Ls**2)
        return 2/self.s * (2 * self.T(0, r0) / (self.b * Ls) * np.power(2 * self.yAPB[phase] * r / (2*self.T(0, r0)), 3/2) - self.beta * xi)

    def APBstrongEdge(self, r, Ls, r0, phase='all'):
        '''
        Anti-phase boundary effect for edge dislocation on strong and shearable particles

        NOTE: Equation from paper would give np.sqrt((np.pi * self.yAPB * r) / (self.V * self.T(np.pi/2)) - 1), but this equation does not follow the plots they have
        '''
        return (2 * self.V * self.T(np.pi/2, r0)) / (np.pi * self.b * Ls) * np.sqrt((np.pi * self.yAPB[phase] * r) / (self.V * self.T(np.pi/2, r0)))

    def APBstrongScrew(self, r, Ls, r0, phase='all'):
        '''
        Anti-phase boundary effect for screw dislocation on strong and shearable particles

        NOTE: Equation from paper would give np.sqrt((np.pi * self.yAPB * r) / (self.V * self.T(np.pi/2)) - 1), but this equation does not follow the plots they have
        '''
        return (2 * self.V * self.T(0, r0)) / (np.pi * self.b * Ls) * np.sqrt((np.pi * self.yAPB[phase] * r) / (self.V * self.T(0, r0)))

    def SFEweakWideEdge(self, r, Ls, r0, phase='all'):
        '''
        Stacking fault energy effect for edge dislocation on weak and shearable particles for wide stacking faults
        '''
        return 2 * self.T(np.pi/2, r0) / (self.b * Ls) * np.power(r * (self.ySFM - self.ySFP[phase]) / self.T(np.pi/2, r0), 3/2)

    def SFEweakWideScrew(self, r, Ls, r0, phase='all'):
        '''
        Stacking fault energy effect for screw dislocation on weak and shearable particles for wide stacking faults
        '''
        return 2 * self.T(0, r0) / (self.b * Ls) * np.power(r * (self.ySFM - self.ySFP[phase]) / self.T(0, r0), 3/2)

    def SFEstrongWide(self, r, Ls, r0, phase='all'):
        '''
        Stacking fault energy effect for edge dislocation on strong and shearable particles for wide stacking faults
        '''
        return self.J * 2 * r * (self.ySFM - self.ySFP[phase]) / (self.b * Ls)

    def SFEweakNarrowEdge(self, r, Ls, r0, phase='all'):
        '''
        Stacking fault energy effect for edge dislocation on weak and shearable particles for narrow stacking faults
        '''
        return (2 * self.T(np.pi/2, r0)) / (self.b * Ls) * np.power((self.ySFM - self.ySFP[phase]) * np.sqrt(self.SFEWeff(np.pi/2, phase) * r - self.SFEWeff(np.pi/2, phase)**2 / 4) / self.T(np.pi/2, r0), 3/2)

    def SFEweakNarrowScrew(self, r, Ls, r0, phase='all'):
        '''
        Stacking fault energy effect for screw dislocation on weak and shearable particles for narrow stacking faults
        '''
        return (2 * self.T(0, r0)) / (self.b * Ls) * np.power((self.ySFM - self.ySFP[phase]) * np.sqrt(self.SFEWeff(0, phase) * r - self.SFEWeff(0, phase)**2 / 4) / self.T(0, r0), 3/2)

    def SFEstrongNarrowEdge(self, r, Ls, r0, phase='all'):
        '''
        Stacking fault energy effect for edge dislocation on strong and shearable particles for narrow stacking faults
        '''
        return self.J * 2 * (self.ySFM - self.ySFP[phase]) * np.sqrt(self.SFEWeff(np.pi/2, phase) * r - self.SFEWeff(np.pi/2, phase)**2 / 4) / (self.b * Ls)

    def SFEstrongNarrowScrew(self, r, Ls, r0, phase='all'):
        '''
        Stacking fault energy effect for screw dislocation on strong and shearable particles for narrow stacking faults
        '''
        return self.J * 2 * (self.ySFM - self.ySFP[phase]) * np.sqrt(self.SFEWeff(0, phase) * r - self.SFEWeff(0, phase)**2 / 4) / (self.b * Ls)

    def interfacialWeakEdge(self, r, Ls, r0, phase='all'):
        '''
        Interfacial energy effect for edge dislocation on weak and shearable particles
        '''
        return 2 * self.T(np.pi/2, r0) / (self.b * Ls) * np.power(self.gamma[phase] * self.b / self.T(np.pi/2, r0), 3/2)

    def interfacialWeakScrew(self, r, Ls, r0, phase='all'):
        '''
        Interfacial energy effect for screw dislocation on weak and shearable particles
        '''
        return 2 * self.T(0, r0) / (self.b * Ls) * np.power(self.gamma[phase] * self.b / self.T(0, r0), 3/2)

    def interfacialStrongOld(self, r, Ls, r0, phase='all'):
        '''
        Interfacial energy effect strong and shearable particles (independent of dislocation type)
        '''
        return self.J * 2 * self.gamma[phase] / Ls

    


import matplotlib.pyplot as plt
import numpy as np

def plot(diffModel, ax = None, plotReference = True, plotElement = None, zScale = 1, *args, **kwargs):
    '''
    Plots composition profile

    Parameters
    ----------
    ax : matplotlib Axes object
        Axis to plot on
    plotReference : bool
        Whether to plot reference element (composition = 1 - sum(composition of rest of elements))
    plotElement : None or str
        Plots single element if it is defined, otherwise, all elements are plotted
    zScale : float
        Scale factor for z-coordinates
    '''
    if ax is None:
        fig, ax = plt.subplots(1,1)

    if not diffModel.isSetup:
        diffModel.setup()

    if plotElement is not None:
        if plotElement not in diffModel.elements and plotElement in diffModel.allElements:
            x = 1 - np.sum(diffModel.x, axis=0)
        else:
            e = diffModel._getElementIndex(plotElement)
            x = diffModel.x[e]
        ax.plot(diffModel.z/zScale, x, *args, **kwargs)
    else:
        if plotReference:
            refE = 1 - np.sum(diffModel.x, axis=0)
            ax.plot(diffModel.z/zScale, refE, label=diffModel.allElements[0], *args, **kwargs)
        for e in range(len(diffModel.elements)):
            ax.plot(diffModel.z/zScale, diffModel.x[e], label=diffModel.elements[e], *args, **kwargs)
        
    ax.set_xlim([diffModel.zlim[0]/zScale, diffModel.zlim[1]/zScale])
    if plotElement is None:
        ax.legend()
    ax.set_xlabel('Distance (m)')
    ax.set_ylabel('Composition (at.%)')

    return ax

def plotTwoAxis(diffModel, Lelements, Relements, zScale = 1, axL = None, axR = None, *args, **kwargs):
    '''
    Plots composition profile with two y-axes

    Parameters
    ----------
    axL : matplotlib Axes object
        Left axis to plot on
    Lelements : list of str
        Elements to plot on left axis
    Relements : list of str
        Elements to plot on right axis
    axR : matplotlib Axes object (optional)
        Right axis to plot on
        If None, then the right axis will be created
    zScale : float
        Scale factor for z-coordinates
    '''
    if axL is None:
        fig, axL = plt.subplots(1,1)

    if not diffModel.isSetup:
        diffModel.setup()

    if type(Lelements) is str:
        Lelements = [Lelements]
    if type(Relements) is str:
        Relements = [Relements]

    ci = 0
    refE = 1 - np.sum(diffModel.x, axis=0)
    if axR is None:
        axR = axL.twinx()
    for e in range(len(Lelements)):
        if Lelements[e] in diffModel.elements:
            eIndex = diffModel._getElementIndex(Lelements[e])
            axL.plot(diffModel.z/zScale, diffModel.x[eIndex], label=diffModel.elements[eIndex], color = 'C' + str(ci), *args, **kwargs)
            ci = ci+1 if ci <= 9 else 0
        elif Lelements[e] in diffModel.allElements:
            axL.plot(diffModel.z/zScale, refE, label=diffModel.allElements[0], color = 'C' + str(ci), *args, **kwargs)
            ci = ci+1 if ci <= 9 else 0
    for e in range(len(Relements)):
        if Relements[e] in diffModel.elements:
            eIndex = diffModel._getElementIndex(Relements[e])
            axR.plot(diffModel.z/zScale, diffModel.x[eIndex], label=diffModel.elements[eIndex], color = 'C' + str(ci), *args, **kwargs)
            ci = ci+1 if ci <= 9 else 0
        elif Relements[e] in diffModel.allElements:
            axR.plot(diffModel.z/zScale, refE, label=diffModel.allElements[0], color = 'C' + str(ci), *args, **kwargs)
            ci = ci+1 if ci <= 9 else 0

    
    axL.set_xlim([diffModel.zlim[0]/zScale, diffModel.zlim[1]/zScale])
    axL.set_xlabel('Distance (m)')
    axL.set_ylabel('Composition (at.%) ' + str(Lelements))
    axR.set_ylabel('Composition (at.%) ' + str(Relements))
    
    lines, labels = axL.get_legend_handles_labels()
    lines2, labels2 = axR.get_legend_handles_labels()
    axR.legend(lines+lines2, labels+labels2, framealpha=1)

    return axL, axR

def plotPhases(diffModel, ax = None, plotPhase = None, zScale = 1, *args, **kwargs):
    '''
    Plots phase fractions over z

    Parameters
    ----------
    ax : matplotlib Axes object
        Axis to plot on
    plotPhase : None or str
        Plots single phase if it is defined, otherwise, all phases are plotted
    zScale : float
        Scale factor for z-coordinates
    '''
    if ax is None:
        fig, ax = plt.subplots(1,1)

    if not diffModel.isSetup:
        diffModel.setup()

    if plotPhase is not None:
        p = diffModel._getPhaseIndex(plotPhase)
        ax.plot(diffModel.z/zScale, diffModel.p[p], *args, **kwargs)
    else:
        for p in range(len(diffModel.phases)):
            ax.plot(diffModel.z/zScale, diffModel.p[p], label=diffModel.phases[p], *args, **kwargs)
    ax.set_xlim([diffModel.zlim[0]/zScale, diffModel.zlim[1]/zScale])
    ax.set_ylim([0, 1])
    ax.set_xlabel('Distance (m)')
    ax.set_ylabel('Phase Fraction')

    if plotPhase is None:
        ax.legend()

    return ax


from numpy.testing import assert_allclose
import numpy as np

from kawin.precipitation import ShapeFactor
from kawin.precipitation import StrainEnergy

import itertools

Rsingle = 2
Rarray = np.linspace(1, 2, 10)

def test_SphericalOutput():
    '''
    Tests output of spherical shape factors given a single radius or list of radii
    Since these factors are constant, all factors should give a scaler value

    The aspectRatio, eqRadius, kineticFactor and thermoFactor methods returns
        a) a scalar for a single radius
        b) a list or corresponding length for a list of radii

    The normalRadii method returns
        a) a list of length 3 for a single radius
        b) an array of size (n x 3) for a list of radii of length n
    '''
    sf = ShapeFactor()
    sf.setSpherical()

    arSingle = sf.aspectRatio(Rsingle)
    arArray = sf.aspectRatio(Rarray)

    eqRsingle = sf.eqRadiusFactor(Rsingle)
    eqRarray = sf.eqRadiusFactor(Rarray)

    radiiSingle = sf.normalRadii(Rsingle)
    radiiArray = sf.normalRadii(Rarray)

    thermoSingle = sf.thermoFactor(Rsingle)
    thermoArray = sf.thermoFactor(Rarray)

    kineticSingle = sf.kineticFactor(Rsingle)
    kineticArray = sf.kineticFactor(Rarray)

    assert np.isscalar(arSingle) or (type(arSingle) == np.ndarray and arSingle.ndim == 0)
    assert arArray.shape == (10,)
    assert np.isscalar(eqRsingle) or (type(eqRsingle) == np.ndarray and eqRsingle.ndim == 0)
    assert eqRarray.shape == (10,)
    assert radiiSingle.shape == (3,)
    assert radiiArray.shape == (10,3)
    assert np.isscalar(thermoSingle) or (type(thermoSingle) == np.ndarray and thermoSingle.ndim == 0)
    assert thermoArray.shape == (10,)
    assert np.isscalar(kineticSingle) or (type(kineticSingle) == np.ndarray and kineticSingle.ndim == 0)
    assert kineticArray.shape == (10,)

def test_NeedleOutput():
    '''
    Tests output of needle shape factors given a single radius or list of radii

    The aspectRatio, eqRadius, kineticFactor and thermoFactor methods returns
        a) a scalar for a single radius
        b) a list or corresponding length for a list of radii

    The normalRadii method returns
        a) a list of length 3 for a single radius
        b) an array of size (n x 3) for a list of radii of length n
    '''
    sf = ShapeFactor()
    sf.setNeedleShape(ar=2)

    arSingle = sf.aspectRatio(Rsingle)
    arArray = sf.aspectRatio(Rarray)

    eqRsingle = sf.eqRadiusFactor(Rsingle)
    eqRarray = sf.eqRadiusFactor(Rarray)

    radiiSingle = sf.normalRadii(Rsingle)
    radiiArray = sf.normalRadii(Rarray)

    thermoSingle = sf.thermoFactor(Rsingle)
    thermoArray = sf.thermoFactor(Rarray)

    kineticSingle = sf.kineticFactor(Rsingle)
    kineticArray = sf.kineticFactor(Rarray)

    assert np.isscalar(arSingle) or (type(arSingle) == np.ndarray and arSingle.ndim == 0)
    assert arArray.shape == (10,)
    assert np.isscalar(eqRsingle) or (type(eqRsingle) == np.ndarray and eqRsingle.ndim == 0)
    assert eqRarray.shape == (10,)
    assert radiiSingle.shape == (3,)
    assert radiiArray.shape == (10,3)
    assert np.isscalar(thermoSingle) or (type(thermoSingle) == np.ndarray and thermoSingle.ndim == 0)
    assert thermoArray.shape == (10,)
    assert np.isscalar(kineticSingle) or (type(kineticSingle) == np.ndarray and kineticSingle.ndim == 0)
    assert kineticArray.shape == (10,)

def test_PlateOutput():
    '''
    Tests output of plate shape factors given a single radius or list of radii

    The aspectRatio, eqRadius, kineticFactor and thermoFactor methods returns
        a) a scalar for a single radius
        b) a list or corresponding length for a list of radii

    The normalRadii method returns
        a) a list of length 3 for a single radius
        b) an array of size (n x 3) for a list of radii of length n
    '''
    sf = ShapeFactor()
    sf.setPlateShape(ar=2)

    arSingle = sf.aspectRatio(Rsingle)
    arArray = sf.aspectRatio(Rarray)

    eqRsingle = sf.eqRadiusFactor(Rsingle)
    eqRarray = sf.eqRadiusFactor(Rarray)

    radiiSingle = sf.normalRadii(Rsingle)
    radiiArray = sf.normalRadii(Rarray)

    thermoSingle = sf.thermoFactor(Rsingle)
    thermoArray = sf.thermoFactor(Rarray)

    kineticSingle = sf.kineticFactor(Rsingle)
    kineticArray = sf.kineticFactor(Rarray)

    assert np.isscalar(arSingle) or (type(arSingle) == np.ndarray and arSingle.ndim == 0)
    assert arArray.shape == (10,)
    assert np.isscalar(eqRsingle) or (type(eqRsingle) == np.ndarray and eqRsingle.ndim == 0)
    assert eqRarray.shape == (10,)
    assert radiiSingle.shape == (3,)
    assert radiiArray.shape == (10,3)
    assert np.isscalar(thermoSingle) or (type(thermoSingle) == np.ndarray and thermoSingle.ndim == 0)
    assert thermoArray.shape == (10,)
    assert np.isscalar(kineticSingle) or (type(kineticSingle) == np.ndarray and kineticSingle.ndim == 0)
    assert kineticArray.shape == (10,)

def test_CuboidalOutput():
    '''
    Tests output of cuboidal shape factors given a single radius or list of radii

    The aspectRatio, eqRadius, kineticFactor and thermoFactor methods returns
        a) a scalar for a single radius
        b) a list or corresponding length for a list of radii

    The normalRadii method returns
        a) a list of length 3 for a single radius
        b) an array of size (n x 3) for a list of radii of length n
    '''
    sf = ShapeFactor()
    sf.setCuboidalShape(ar=2)

    arSingle = sf.aspectRatio(Rsingle)
    arArray = sf.aspectRatio(Rarray)

    eqRsingle = sf.eqRadiusFactor(Rsingle)
    eqRarray = sf.eqRadiusFactor(Rarray)

    radiiSingle = sf.normalRadii(Rsingle)
    radiiArray = sf.normalRadii(Rarray)

    thermoSingle = sf.thermoFactor(Rsingle)
    thermoArray = sf.thermoFactor(Rarray)

    kineticSingle = sf.kineticFactor(Rsingle)
    kineticArray = sf.kineticFactor(Rarray)

    assert np.isscalar(arSingle) or (type(arSingle) == np.ndarray and arSingle.ndim == 0)
    assert arArray.shape == (10,)
    assert np.isscalar(eqRsingle) or (type(eqRsingle) == np.ndarray and eqRsingle.ndim == 0)
    assert eqRarray.shape == (10,)
    assert radiiSingle.shape == (3,)
    assert radiiArray.shape == (10,3)
    assert np.isscalar(thermoSingle) or (type(thermoSingle) == np.ndarray and thermoSingle.ndim == 0)
    assert thermoArray.shape == (10,)
    assert np.isscalar(kineticSingle) or (type(kineticSingle) == np.ndarray and kineticSingle.ndim == 0)
    assert kineticArray.shape == (10,)

def test_AspectRatio():
    '''
    Tests that aspect ratio can be as a function
    '''
    arFunc = lambda R : 0.5 * R + 3
    sf = ShapeFactor()
    sf.setAspectRatio(arFunc)
    Rarray = sf.aspectRatio(np.array([1, 2]))

    assert_allclose(Rarray, np.array([3.5, 4]), rtol=1e-5)

def test_StrainOutput():
    '''
    Tests that strain energy outputs:
        a) a scalar for a single set of radii (3 length array)
        b) a list for a list of radii (n x 3 array)
    '''
    se = StrainEnergy()
    se.setEigenstrain(0.01)
    se.setElasticConstants(150e9, 100e9, 75e9)

    rSingle = np.random.random(3)
    rArray = np.random.random((10, 3))

    elSingle = se.strainEnergy(rSingle)
    elArray = se.strainEnergy(rArray)

    assert np.isscalar(elSingle) or (type(elSingle) == np.ndarray and elSingle.ndim == 0)
    assert elArray.shape == (10,)

def test_StrainValues():
    '''
    Test strain energy calculation of arbitrary system

    Parameters are taken from 11_Extra_Factors for the Cu-Ti system
    '''
    se = StrainEnergy()
    se.setEllipsoidal()
    se.setElasticConstants(168.4e9, 121.4e9, 75.4e9)
    se.setEigenstrain([0.022, 0.022, 0.003])
    se.setup()

    aspect = 1.5
    rSph = 4e-9 / np.cbrt(aspect)
    r = np.array([rSph, rSph, aspect*rSph])
    E = se.strainEnergy(r)

    assert_allclose(E, 1.22956765e-17, rtol=1e-3)

def test_AspectRatio():
    '''
    Test eq aspect ratio calculation of arbitrary system

    Parameters are taken from 11_Extra_Factors for the IN718 system
    '''
    se = StrainEnergy()
    se.setEigenstrain([6.67e-3, 6.67e-3, 2.86e-2])
    se.setModuli(G=57.1e9, nu=0.33)
    se.setEllipsoidal()
    se.setup()

    sf = ShapeFactor()
    sf.setPlateShape()

    gamma = 0.02375
    Rsph = np.array([5e-10])
    eqAR = se.eqAR_bySearch(Rsph, gamma, sf)
    R = 2*Rsph*eqAR / np.cbrt(eqAR**2)

    assert_allclose(R, [1.13444719e-9], rtol=1e-3)
    assert_allclose(eqAR, [1.46], rtol=1e-3)

def test_different_strain_energy_inputs():
    '''
    Make sure the elastic tensor is the same for different types of inputs
    
    Following options in kawin are:
        2nd rank elastic tensor (6x6 matrix)
        Elastic constants c11, c12 and c44
        2 different moduli (from E, nu, G, lambda, K, or M)

    This will use the G and nu parameters from 11_Extra_Factors for the IN718 example
        Any values should work though since we're just checking that the elastic tensor is the same
    '''
    G = 57.1e9                      #Shear modulus
    nu = 0.33                       #Poisson ratio
    E = 2*G*(1+nu)                  #Elastic modulus
    lam = 2*G*nu / (1-2*nu)         #Lame's first parameter
    K = 2*G*(1+nu)/(3*(1-2*nu))     #Bulk modulus
    M = 2*G*(1-nu)/(1-2*nu)         #P-wave modulus

    c11 = E*(1-nu)/((1+nu)*(1-2*nu))
    c12 = E*nu/((1+nu)*(1-2*nu))
    c44 = G

    se = StrainEnergy()

    r2Tensor = np.array([[c11, c12, c12, 0, 0, 0], [c12, c11, c12, 0, 0, 0], [c12, c12, c11, 0, 0, 0], [0, 0, 0, c44, 0, 0], [0, 0, 0, 0, c44, 0], [0, 0, 0, 0, 0, c44]])
    r4Tensor = se._convert2To4rankTensor(r2Tensor)
    
    #Test 2nd rank tensor input
    se.setElasticTensor(r2Tensor)
    assert_allclose(se._c4, r4Tensor, rtol=1e-3)

    #Test elastic constants input
    se.setElasticConstants(c11, c12, c44)
    assert_allclose(se._c4, r4Tensor, rtol=1e-3)

    #This is in the order of the if statements in StrainEnergy._setModuli so it's easier to debug
    moduli = {'E': E, 'nu': nu, 'G': G, 'lam': lam, 'K': K, 'M': M}
    moduli_names = moduli.keys()

    #Test each pair of moduli as inputs
    for pair in itertools.combinations(moduli_names, 2):
        moduli_input = {m: moduli[m] for m in pair}
        se.setModuli(**moduli_input)
        assert_allclose(se._c4, r4Tensor, rtol=1e-3)









from kawin.precipitation.coupling import StrengthModel
import numpy as np

sm = StrengthModel()
G, b, nu = 79.3e9, 0.25e-9, 1/3
bp, ri = b, 2*b
eps, Gp = 0.001, 70e9
yAPB, ySFM, ySFP, gamma = 0.04, 0.1, 0.05, 0.5
sm.setDislocationParameters(G, b, nu, ri, theta=90, psi=120)
sm.setCoherencyParameters(eps)
sm.setModulusParameters(Gp, phase='all')
sm.setModulusParameters(2*Gp, phase='beta')
sm.setAPBParameters(yAPB, phase='alpha')
sm.setSFEParameters(ySFM, ySFP, bp, phase='beta')
sm.setInterfacialParameters(gamma)
sm.setTaylorFactor(1)

sm2 = StrengthModel()
sm2.setDislocationParameters(G, b, nu, ri, theta=90, psi=120)
sm2.setCoherencyParameters(eps, phase='alpha')

rs = 50e-9
Lss = 300e-9 - 2*rs

def test_coherency_edge():
    '''
    Tests weak and strong coherency contribution for edge dislocations
    '''
    weak = sm.coherencyWeak(rs, Lss, Lss / np.sqrt(np.cos(sm.psi / 2)))
    strong = sm.coherencyStrong(rs, Lss, Lss)
    assert(np.allclose([weak, strong], [7.26102e7, 4.82891e7], atol=0, rtol=1e-3))

def test_coherency_screw():
    '''
    Tests weak and strong coherency contribution for screw dislocations
    '''
    sm.setDislocationParameters(G, b, nu, ri, theta=0, psi=120)
    weak = sm.coherencyWeak(rs, Lss, Lss / np.sqrt(np.cos(sm.psi / 2)))
    strong = sm.coherencyStrong(rs, Lss, Lss)
    sm.setDislocationParameters(G, b, nu, ri, theta=90, psi=120)
    assert(np.allclose([weak, strong], [1.18431e7, 1.27934e8], atol=0, rtol=1e-3))

def test_modulus_edge():
    '''
    Tests weak and strong modulus contribution for edge dislocations
    '''
    weak = sm.modulusWeak(rs, Lss, Lss / np.sqrt(np.cos(sm.psi / 2)))
    strong = sm.modulusStrong(rs, Lss, Lss)
    assert(np.allclose([weak, strong], [5.38138e7, 5.25096e7], atol=0, rtol=1e-3))

def test_modulus_screw():
    '''
    Tests weak and strong modulus contribution for screw dislocations
    '''
    sm.setDislocationParameters(G, b, nu, ri, theta=0, psi=120)
    weak = sm.modulusWeak(rs, Lss, Lss / np.sqrt(np.cos(sm.psi / 2)))
    strong = sm.modulusStrong(rs, Lss, Lss)
    sm.setDislocationParameters(G, b, nu, ri, theta=90, psi=120)
    assert(np.allclose([weak, strong], [2.69069e7, 5.25096e7], atol=0, rtol=1e-3))

def test_APB_edge():
    '''
    Tests weak and strong anti-phase boundary contribution for edge dislocations
    '''
    weak = sm.APBweak(rs, Lss, Lss / np.sqrt(np.cos(sm.psi / 2)), phase='alpha')
    strong = sm.APBstrong(rs, Lss, Lss, phase='alpha')
    assert(np.allclose([weak, strong], [8.42212e7, 5.79670e7], atol=0, rtol=1e-3))

def test_APB_screw():
    '''
    Tests weak and strong anti-phase boundary contribution for screw dislocations
    '''
    sm.setDislocationParameters(G, b, nu, ri, theta=0, psi=120)
    weak = sm.APBweak(rs, Lss, Lss / np.sqrt(np.cos(sm.psi / 2)), phase='alpha')
    strong = sm.APBstrong(rs, Lss, Lss, phase='alpha')
    sm.setDislocationParameters(G, b, nu, ri, theta=90, psi=120)
    assert(np.allclose([weak, strong], [3.36224e7, 1.15934e8], atol=0, rtol=1e-3))

def test_SFE_edge():
    '''
    Tests weak and strong stacking fault energy contribution for edge dislocations
    '''
    weak = sm.SFEweak(rs, Lss, Lss / np.sqrt(np.cos(sm.psi / 2)), phase='beta')
    strong = sm.SFEstrong(rs, Lss, Lss, phase='beta')
    assert(np.allclose([weak, strong], [3.83623e7, 4.19031e7], atol=0, rtol=1e-3))

def test_SFE_screw():
    '''
    Tests weak and strong stacking fault energy contribution for screw dislocations
    '''
    sm.setDislocationParameters(G, b, nu, ri, theta=0, psi=120)
    weak = sm.SFEweak(rs, Lss, Lss / np.sqrt(np.cos(sm.psi / 2)), phase='beta')
    strong = sm.SFEstrong(rs, Lss, Lss, phase='beta')
    sm.setDislocationParameters(G, b, nu, ri, theta=90, psi=120)
    assert(np.allclose([weak, strong], [1.03693e7, 2.78075e7], atol=0, rtol=1e-3))

def test_IFE_edge():
    '''
    Tests weak and strong interfacial energy contribution for edge dislocations
    '''
    weak = sm.interfacialWeak(rs, Lss, Lss / np.sqrt(np.cos(sm.psi / 2)))
    strong = sm.interfacialStrong(rs, Lss, Lss)
    assert(np.allclose([weak, strong], [1.58121e6, 5.00000e6], atol=0, rtol=1e-3))

def test_IFE_screw():
    '''
    Tests weak and strong interfacial energy contribution for screw dislocations
    '''
    sm.setDislocationParameters(G, b, nu, ri, theta=0, psi=120)
    weak = sm.interfacialWeak(rs, Lss, Lss / np.sqrt(np.cos(sm.psi / 2)))
    strong = sm.interfacialStrong(rs, Lss, Lss)
    sm.setDislocationParameters(G, b, nu, ri, theta=90, psi=120)
    assert(np.allclose([weak, strong], [7.90607e5, 5.00000e6], atol=0, rtol=1e-3))

def test_orowan():
    '''
    Tests orowan contribution
    '''
    oro = sm.orowan(rs, Lss)
    assert(np.allclose(oro, 1.02373e8, atol=0, rtol=1e-3))

def test_strength_output():
    '''
    Tests strength output values for when not all contributions are supplied
    No phase input or 'all' input should apply contributions to all phases
    Having a specific phase as an input should apply contributions to only that phase

    'all' should include only coherency, modulus and interfacial energy
    'alpha' should include coherency, modulus, anti-phase boundary and interfacial energy
    'beta' should include coherency, modulus, stacking fault energy and interfacial energy
    '''
    rs = 50e-9
    Ls = 300e-9 - 2*rs
    Leff = Ls / np.sqrt(np.cos(sm.psi / 2))

    cohWeak = sm.coherencyWeak(rs, Ls, Leff)
    cohStrong = sm.coherencyStrong(rs, Ls, Ls)

    #Values for 'all' and 'beta' will be defined separately to make sure getting strength
    #for the beta phase will override 'all'
    modWeak = sm.modulusWeak(rs, Ls, Leff)
    modStrong = sm.modulusStrong(rs, Ls, Ls)
    modWeakBeta = sm.modulusWeak(rs, Ls, Leff, 'beta')
    modStrongBeta = sm.modulusStrong(rs, Ls, Ls, 'beta')
    apbWeak = sm.APBweak(rs, Ls, Leff, 'alpha')
    apbStrong = sm.APBstrong(rs, Ls, Ls, 'alpha')
    sfeWeak = sm.SFEweak(rs, Ls, Leff, 'beta')
    sfeStrong = sm.SFEstrong(rs, Ls, Ls, 'beta')
    ifeWeak = sm.interfacialWeak(rs, Ls, Leff)
    ifeStrong = sm.interfacialStrong(rs, Ls, Ls)
    orowan = sm.orowan(rs, Ls)


    tauall = sm.getStrengthContributions(rs, Ls, )
    taualpha = sm.getStrengthContributions(rs, Ls, 'alpha')
    taubeta = sm.getStrengthContributions(rs, Ls, 'beta')

    assert(np.allclose(tauall[0], [cohWeak, modWeak, ifeWeak], atol=0, rtol=1e-3))
    assert(np.allclose(tauall[1], [cohStrong, modStrong, ifeStrong], atol=0, rtol=1e-3))
    assert(np.allclose(tauall[2], [orowan], atol=0, rtol=1e-3))

    assert(np.allclose(taualpha[0], [cohWeak, modWeak, apbWeak, ifeWeak], atol=0, rtol=1e-3))
    assert(np.allclose(taualpha[1], [cohStrong, modStrong, apbStrong, ifeStrong], atol=0, rtol=1e-3))
    assert(np.allclose(taualpha[2], [orowan], atol=0, rtol=1e-3))

    assert(np.allclose(taubeta[0], [cohWeak, modWeakBeta, sfeWeak, ifeWeak], atol=0, rtol=1e-3))
    assert(np.allclose(taubeta[1], [cohStrong, modStrongBeta, sfeStrong, ifeStrong], atol=0, rtol=1e-3))
    assert(np.allclose(taubeta[2], [orowan], atol=0, rtol=1e-3))

def test_no_strength_contribution():
    '''
    Tests that strength model will return 0 if there is no strength contribution

    Also test that the model will return an array of 0s if the input is an array
    '''
    rs = 50e-9
    Ls = 300e-9 - 2*rs
    Leff = Ls / np.sqrt(np.cos(sm.psi / 2))

    taugamma = sm2.getStrengthContributions(rs, Ls, 'gamma')
    strengthgamma = sm2.combineStrengthContributions(taugamma[0], taugamma[1], taugamma[2])

    N = 5
    rs = np.linspace(10e-9, 50e-9, N)
    Ls = 300e-9 - 2*rs
    Leff = Ls / np.sqrt(np.cos(sm.psi / 2))

    taugamma2 = sm2.getStrengthContributions(rs, Ls, 'gamma')
    strengthgamma2 = sm2.combineStrengthContributions(taugamma[0], taugamma[1], taugamma[2])


    assert(np.allclose(strengthgamma, 0, atol=0, rtol=1e-3))
    assert(np.allclose(strengthgamma2, np.zeros(N), atol=0, rtol=1e-3))


from enum import Enum

'''
Defines class to handle a single stopping conditions

Per iteration, these will take in a model, and check with internal members to see if stopping condition has been satisfied
If it has, then it will be set to True and the time will be recorded
These can also be checked if they were satisfied already if we want to use them to stop a simulation

TODO: Abstract out the stopping condition so it can be used in GenericModel
'''

class Inequality (Enum):
    GREATER_THAN = 0
    LESSER_THAN = 1

class PrecipitationStoppingCondition:
    '''
    Parameters
    ----------
    condition : Inequality enum
        GREATER_THAN -> result > value
        LESS_THAN -> result < value
    value : double
    phase : str
    element : el
    '''
    def __init__(self, condition, value, phase = None, element = None):
        self._condition = condition
        self._value = value
        self._isSatisfied = False
        self._satisfiedTime = -1
        self._phase = phase
        self._element = element
        self._modelVar = None

    def reset(self):
        '''
        Resets condition to being not yet satisfied
        '''
        self._isSatisfied = False
        self._satisfiedTime = -1

    def _poll(self, model, n):
        '''
        Gets current value of attribute at iteration n for phase p

        Parameters
        ----------
        model : PrecipitateModel
        n : int
            Iteration number

        Returns value (float) of attribute at n,p
        '''
        p = model.phaseIndex(self._phase)
        return getattr(model, self._modelVar)[n,p]
    
    def _testCondition(self, model):
        '''
        Private function only testing if stopping condition is satisfied based off current state of model

        Parameters
        ----------
        model : PrecipitateModel

        Returns bool for whether condition is satisfied or not
        '''
        if self._condition == Inequality.GREATER_THAN:
            return self._poll(model, model.n) > self._value
        else:
            return self._poll(model, model.n) < self._value
    
    def testCondition(self, model):
        '''
        Tests if condition is satisfied, if so, then interpolate to find time when it was satisfied

        Parameters
        ----------
        model : PrecipitateModel
        '''
        if not self._isSatisfied:
            self._isSatisfied = self._testCondition(model)

            if self._isSatisfied:
                if model.n > 0:
                    currVal, currTime = self._poll(model, model.n), model.time[model.n]
                    prevVal, prevTime = self._poll(model, model.n-1), model.time[model.n-1]
                    self._satisfiedTime = (currTime - prevTime) * (self._value - prevVal) / (currVal - prevVal) + prevTime
                else:
                    self._satisfiedTime = model.time[model.n]

    def isSatisfied(self):
        '''
        Returns whether condition is satisfied
        '''
        return self._isSatisfied
    
    def satisfiedTime(self):
        '''
        Returns time when condition was satisfied
        '''
        return self._satisfiedTime

class VolumeFractionCondition (PrecipitationStoppingCondition):
    def __init__(self, condition, value, phase = None):
        super().__init__(condition, value, phase = phase)
        self._modelVar = 'betaFrac'

class AverageRadiusCondition (PrecipitationStoppingCondition):
    def __init__(self, condition, value, phase = None):
        super().__init__(condition, value, phase = phase)
        self._modelVar = 'avgR'
        
class DrivingForceCondition (PrecipitationStoppingCondition):
    def __init__(self, condition, value, phase = None):
        super().__init__(condition, value, phase = phase)
        self._modelVar = 'dGs'

class NucleationRateCondition (PrecipitationStoppingCondition):
    def __init__(self, condition, value, phase = None):
        super().__init__(condition, value, phase = phase)
        self._modelVar = 'nucRate'

class PrecipitateDensityCondition (PrecipitationStoppingCondition):
    def __init__(self, condition, value, phase = None):
        super().__init__(condition, value, phase = phase)
        self._modelVar = 'precipitateDensity'

class CompositionCondition (PrecipitationStoppingCondition):
    def __init__(self, condition, value, element = None):
        super().__init__(condition, value, element = element)
        self._modelVar = 'xComp'

    def _poll(self, model, n):
        e = 0 if self._element is None else model.elements.index(self._element)
        return getattr(model, self._modelVar)[n,e]



from kawin.precipitation import PrecipitateModel, VolumeParameter
from kawin.diffusion import SinglePhaseModel
from kawin.thermo import BinaryThermodynamics, MulticomponentThermodynamics
from kawin.GenericModel import GenericModel, Coupler
from kawin.solver import SolverType
import numpy as np
from numpy.testing import assert_allclose
from kawin.tests.datasets import *

AlZrTherm = BinaryThermodynamics(ALZR_TDB, ['AL', 'ZR'], ['FCC_A1', 'AL3ZR'], drivingForceMethod='tangent')
NiAlCrTherm = MulticomponentThermodynamics(NICRAL_TDB, ['NI', 'AL', 'CR'], ['FCC_A1', 'FCC_L12'], drivingForceMethod='tangent')

AlZrTherm.setDFSamplingDensity(2000)
AlZrTherm.setEQSamplingDensity(500)
NiAlCrTherm.setDFSamplingDensity(2000)
NiAlCrTherm.setEQSamplingDensity(500)

def test_iterators():
    '''
    Tests explicit euler and RK4 iterators
    '''
    class TestModel(GenericModel):
        def __init__(self):
            self.reset()

        def reset(self):
            self.x = np.array([0])
            self.time = np.zeros(1)

        def getCurrentX(self):
            return self.time[-1], [self.x[-1]]
        
        def getdXdt(self, t, x):
            return [np.cos(t)]
        
        def getDt(self, dXdt):
            return 0.001
        
        def postProcess(self, time, x):
            self.time = np.append(self.time, time)
            self.x = np.append(self.x, x[0])
            return x, False
        
    m = TestModel()
    m.solve(10, solverType=SolverType.EXPLICITEULER)
    eulerX = m.x[-1]

    m.reset()
    m.solve(10, solverType=SolverType.RK4)
    rkX = m.x[-1]

    assert_allclose(eulerX, np.sin(10), rtol=1e-2)
    assert_allclose(rkX, np.sin(10), rtol=1e-2)

def test_coupler_shape():
    '''
    Test that coupler returns correct shape when flattening and unflattening arrays

    Here we use a precipitate model and diffusion model where the shape of x is:
        Precipitate model: [(bins,)]
        Diffusion model: [(elements,cells,)]
    Flattening the arrays will result in a 1D array of [bins + elements*cells]
    '''
    #Create model
    p_model = PrecipitateModel()
    bins = 75
    minBins = 50
    maxBins = 100
    p_model.setPBMParameters(cMin=1e-10, cMax=1e-8, bins=bins, minBins=minBins, maxBins=maxBins)

    xInit = 4e-3        #Initial composition (mole fraction)
    p_model.setInitialComposition(xInit)

    T = 450 + 273.15    #Temperature (K)
    p_model.setTemperature(T)

    gamma = 0.1         #Interfacial energy (J/m2)
    p_model.setInterfacialEnergy(gamma)

    D0 = 0.0768         #Diffusivity pre-factor (m2/s)
    Q = 242000          #Activation energy (J/mol)
    Diff = lambda x, T: D0 * np.exp(-Q / (8.314 * T))
    p_model.setDiffusivity(Diff)

    a = 0.405e-9        #Lattice parameter
    Va = a**3           #Atomic volume of FCC-Al
    Vb = a**3           #Assume Al3Zr has same unit volume as FCC-Al
    atomsPerCell = 4    #Atoms in an FCC unit cell
    p_model.setVolumeAlpha(Va, VolumeParameter.ATOMIC_VOLUME, atomsPerCell)
    p_model.setVolumeBeta(Vb, VolumeParameter.ATOMIC_VOLUME, atomsPerCell)

    #Average grain size (um) and dislocation density (1e15)
    p_model.setNucleationDensity(grainSize = 1, dislocationDensity = 1e15)
    p_model.setNucleationSite('dislocations')

    #Set thermodynamic functions
    p_model.setThermodynamics(AlZrTherm, addDiffusivity=False)

    #Define mesh spanning between -1mm to 1mm with 50 volume elements
    #Since we defined L12, the disordered phase as DIS_ attached to the front
    N = 20
    d_model = SinglePhaseModel([-1e-3, 1e-3], N, ['NI', 'AL', 'CR'], ['DIS_FCC_A1'])

    #Define Cr and Al composition, with step-wise change at z=0
    d_model.setCompositionLinear(0.077, 0.359, 'CR')
    d_model.setCompositionLinear(0.054, 0.062, 'AL')

    d_model.setThermodynamics(NiAlCrTherm)
    d_model.setTemperature(1200 + 273.15)

    coupled_model = Coupler([p_model, d_model])
    coupled_model.setup()

    t, x = coupled_model.getCurrentX()
    x_flat = coupled_model.flattenX(x)
    x_restore = coupled_model.unflattenX(x_flat, x)

    assert(len(x) == 2)
    assert(len(x[0]) == 1 and x[0][0].shape == (bins,))
    assert(len(x[1]) == 1 and x[1][0].shape == (2,N))
    assert(x_flat.shape == (bins+2*N,))
    assert(len(x_restore) == len(x))
    assert(len(x_restore[0]) == len(x[0]) and x_restore[0][0].shape == x[0][0].shape)
    assert(len(x_restore[1]) == 1 and x_restore[1][0].shape == x[1][0].shape)


from kawin.precipitation import PrecipitateModel
from kawin.diffusion.Diffusion import DiffusionModel
import matplotlib.pyplot as plt
import numpy as np

def test_precipitate_plotting():
    binary_single = PrecipitateModel(phases=['beta'], elements=['A'])
    binary_multi = PrecipitateModel(phases=['beta', 'gamma', 'zeta'], elements=['A'])
    ternary_single = PrecipitateModel(phases=['beta'], elements=['A', 'B'])
    ternary_multi = PrecipitateModel(phases=['beta', 'gamma', 'zeta'], elements=['A', 'B'])

    models = [
        (binary_single, 1, 1),
        (binary_multi, 1, 3),
        (ternary_single, 2, 1),
        (ternary_multi, 2, 3),
    ]

    varTypes = [
        ('Volume Fraction', [2]),
        ('Total Volume Fraction', None),
        ('Critical Radius', [2]),
        ('Average Radius', [2]),
        ('Volume Average Radius', [2]),
        ('Total Average Radius', None),
        ('Total Volume Average Radius', None),
        ('Aspect Ratio', [2]),
        ('Total Aspect Ratio', None),
        ('Driving Force', [2]),
        ('Nucleation Rate', [2]),
        ('Total Nucleation Rate', None),
        ('Precipitate Density', [2]),
        ('Total Precipitate Density', None),
        ('Temperature', None),
        ('Composition', [1]),
        ('Eq Composition Alpha', [1,2]),
        ('Eq Composition Beta', [1,2]),
        ('Supersaturation', [2]),
        ('Eq Volume Fraction', [2]),
        ('Size Distribution', [2]),
        ('Size Distribution Curve', [2]),
        ('Size Distribution KDE', [2]),
        ('Size Distribution Density', [2]),
    ]

    for m in models:
        for v in varTypes:
            fig, ax = plt.subplots(1,1)
            m[0].plot(ax, v[0])
            numLines = len(ax.lines)
            plt.close(fig)

            #Check that the number of lines on the plot correspond to the right amount
            #   Number of lines should either be 1, elements, phases or elements*phases depending on variable
            desiredNumber = 1
            if v[1] is not None:
                desiredNumber = np.prod([m[vi] for vi in v[1]], dtype=np.int32)
            assert numLines == desiredNumber

def test_diffusion_plotting():
    #Single phase and Homogenizaton model goes through the same path for plotting
    binary_single = DiffusionModel(zlim=[-1,1], N=100, elements=['A', 'B'], phases=['alpha'])
    binary_multi = DiffusionModel(zlim=[-1,1], N=100, elements=['A', 'B'], phases=['alpha', 'beta', 'gamma'])
    ternary_single = DiffusionModel(zlim=[-1,1], N=100, elements=['A', 'B', 'C'], phases=['alpha'])
    ternary_multi = DiffusionModel(zlim=[-1,1], N=100, elements=['A', 'B', 'C'], phases=['alpha', 'beta', 'gamma'])

    models = [
        (binary_single, 2, 1),
        (binary_multi, 2, 3),
        (ternary_single, 3, 1),
        (ternary_multi, 3, 3),
    ]

    for m in models:
        m[0].setTemperature(900)

        #For each plot, check that the number of lines correspond to number of elements or phases
        #For 'plot', number of lines should be elements (with or without reference) or a single element
        #For 'plotTwoAxis', number of lines for each axis should be length of input array
        #For 'plotPhases', number of lines is number of phases or single phase
        fig, ax = plt.subplots(1,1)
        m[0].plot(ax, plotReference = False)
        assert len(ax.lines) == m[1]-1
        plt.close(fig)

        fig, ax = plt.subplots(1,1)
        m[0].plot(ax, plotReference = True)
        assert len(ax.lines) == m[1]
        plt.close(fig)

        fig, ax = plt.subplots(1,1)
        m[0].plot(ax, plotElement = m[0].allElements[0])
        assert len(ax.lines) == 1
        plt.close(fig)

        fig, ax = plt.subplots(1,1)
        m[0].plot(ax, plotElement = m[0].allElements[1])
        assert len(ax.lines) == 1
        plt.close(fig)


        fig, axL = plt.subplots(1,1)
        axR = ax.twinx()
        m[0].plotTwoAxis(Lelements=[m[0].allElements[0]], Relements = m[0].allElements[1:], axL=axL, axR=axR)
        assert len(axL.lines) == 1
        assert len(axR.lines) == len(m[0].allElements)-1
        plt.close(fig)

        fig, ax = plt.subplots(1,1)
        m[0].plotPhases(ax)
        assert len(ax.lines) == m[2]
        plt.close(fig)

        fig, ax = plt.subplots(1,1)
        m[0].plotPhases(ax, plotPhase=m[0].phases[0])
        assert len(ax.lines) == 1
        plt.close(fig)